--[[
	DuckMod - A World of Warcraft add-on library
	Copyright (C) 2009-2012  Dag Bakken

	DuckMod is free software: you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation, either version 3 of the License, or
	(at your option) any later version.

	DuckMod is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with DuckMod.  If not, see <http://www.gnu.org/licenses/>.
]]

--[[
	isConnected = UnitIsConnected("unit")
		unit - A Unit to query ( string , unitID )

]]



TestDebugTable={}

local TV=2.9906;
-- 3.0 We're aiming for this release
-- 2.9910
-- - Inter-guild sync
-- 2.9906
-- - Alpha implementation of Ayla script, prune, new/update sync flag, DuckModDebug added,
--   database soft link
-- 2.9905
-- + Full local sync implemented
-- 2.9904
-- + Database list login sync
-- 2.9903
-- + Renderer: Deprecated in-text links as those widgets measure wrong
-- +		   Capture clicks in the canvas, find widget that was clicked, show link menu
-- 2.9902
-- + Database: Beta (All functionality initial testing done, bigger scale testing imminent)
-- 2.9901
-- + Database: alpha
--   All functionality present, but untested.
-- 2.10
-- + Removed need for GetTime after it broke in patch 4.3
-- 2.0802
-- + Rendering reset to top of page at Render(), bug in inheritSys fixed
-- 2.0801
-- + Further work on wownet and rendering
-- 2.08
-- + Further implementation of WoWnet
-- 2.0703
-- + Rendering speed-up
-- 2.0702
-- 2.0701
-- + Various minor changes, and version change for conflict detection
-- 2.07
-- + WoWnet v0.01: Data transmission by Base64 coding for full digital
--   transmission. Supports multiple page reception.
-- o WN Renderer
-- 2.06
-- + Work on WoWnet progresses. Finally.
-- + Added simple generic multi-threading library methods
-- + Fixed bug in cache that effectively switched it off
-- 2.04
-- + Changed cache to only work for decoded base data (who), enabling
--   decode of many sources without corrupting the cache.
-- + Added an extra level for cache tables
-- 2.0204
-- + Detects uncoded text in "Decode"
-- + Added "Purge" for table-cache
-- 2.0203
-- + Fixed showstopper bug in event handling
-- + Fixed bug in password and coding
-- 2.0202
-- + Table compression enhancement
-- 2.0201
-- + Fix for OnUpdate
-- 2.02
-- + table (de)compression system
-- + ^^ Cache
-- 2.0001
-- + password protection on text
-- + modified base64 coding functions


local DUCKNET_VERSIONTEXT = "DuckNet v1.09";
local DUCKNET_VERSION_DATA = "003";	-- The data structure in use
local DUCKNET_VERSION_API = 2;
local DUCKNET_FROMCODE = "|";
local DUCKNET_TOCODE = "\1";			-- This MUST be only one character
local DUCKNET_DATA="\2";				-- NEW, old="!"
local DUCKNET_ENTRY="\3";				-- NEW, old="#"
local DUCKNET_COMMAND="\4";				-- NEW, old=":"
local DUCKNET_FREESPLIT1="\5";			-- Free splitter for addons
local DUCKNET_QUOTATION="\6";			-- NEW, old="\""
local DUCKNET_WNSPLIT1="\7";			-- WoWnet splitter 1
local DUCKNET_TREESEP="\8";				-- Tree separator
local DUCKNET_DISCONNECT="\10";			-- Never use this
local DUCKNET_LINEBREAKnative = "\n";
local DUCKNET_LINEBREAKhtml = "<BR/>";
local DUCKNET_ENTRYBREAKCONTINUE = "<EBC/>";
local DUCKNET_BOOLVALDATA="BoolVal:";

local DUCKNET_NEG_STOPPED = -1;
local DUCKNET_NEG_RUNNING = 0;
local DUCKNET_NEG_DONE = 5;

local DUCKNET_HEARTBEAT = 0.2;
local DUCKNET_TRANS_STOP = 0;
local DUCKNET_TRANS_START = 1;

DUCKNET_REQ_NEWDATA = "DuckNet-Req-New-Data";
DUCKNET_REQ_RETRANSMIT = "DuckNet-Req-Retransmit-Data";
DUCKNET_ACT_TRANSMIT = "DuckNet-Act-Transmit-Data";
DUCKNET_ACT_TRANSMITDONE = "DuckNet-Act-Transmit-Data-Done";
DUCKNET_ACT_MYSTAMP = "DuckNet-Act-MyStamp";


if (not DuckMod) then DuckMod = {}; end			-- Make it if empty


if (not DuckMod[TV]) then					-- Don't mess with it if it's loaded

-- Helper colours
DuckMod[TV]={								-- Add to array
	Color = {
		White="|cFFFFFFFF",
		Green="|cFF00FF00",
		Red="|cFFFF0000",
		lBlue="|cFF6060FF",				-- Light blue
		hBlue="|cFEA0A0FF",				-- Highlight blue
		Yellow="|cFFFFFF00",
	},
	Talk={},
};

local DM=DuckMod[TV];
globalDMROOT=DM;
local dbg=nil;
DM.SentBlindWhisperListOfReceivers={};

-- Make a random number of integer size with fractional part
function DM:Random(bottom,top)
	if (not bottom) then
		bottom=0;
		top=1;
	elseif (not top) then
		top=bottom;
		bottom=0;
	end

	local num=math.random();		-- Generate a fractional number
	num=num*(top-bottom);			-- Set span
	num=num+bottom;					-- Set floor
	return num;
end

-- Global time-keeping
--function DM:Time()
--	if (self.TimeOffset) then return time()-self.TimeOffset; end
--	return time();
--end

--function DM:TimeDiff(diff)
--	if (not diff) then
--		-- Get data
--		local h=GetGameTime();		-- Hours
--		local t=date("*t");			-- Computer-date
--		diff=t.hour-h;				-- - Diff
--		if (diff>11) then diff=diff-24;
--		elseif (diff<-11) then diff=diff+24;
--		end
--	end
--	self.TimeOffset=diff*3600;
----	("Time-offset to subtract is: "..diff,0,1,1);
--end

-- Table size
function DM:Length(t)
	if (not t) then return 0; end
	local count=0;
	for _ in pairs(t) do count=count+1; end
	return count;
end

-- Safe table-copy with optional merge (equal entries will be overwritten, arg1 has pri)
function DM:CopyTable(t,new)
	if (not t and not new) then return nil; end
	if (not t) then t={}; end
	if (not new) then new={}; end
	local i,v;
	for i,v in pairs(t) do
		if (type(v)=="table") then new[i]=self:CopyTable(v,new[i]); else new[i]=v; end
	end
	return new;
end

-- Safe table-clear
function DM:ClearTable(t)
	for k in pairs(t) do t[k]=nil; end;
end

-- Return the common DuckMod ID
function DM:GetID(link)
	if (type(link)~="string") then
		if (type(link)~="number") then return nil,nil; end
		_,link=GetItemInfo(link);
	end
	local itemID,i1,i2,i3,i4,i5,i6=link:match("|Hitem:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)|h%[(.-)%]|h");
	if (not i6) then
		itemID,i1,i2,i3,i4,i5,i6=link:match("item:(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+):(%p?%d+)");
	end
	if (not i6) then return; end
	return "item:"..itemID..":"..i1..":"..i2..":"..i3..":"..i4..":"..i5..":"..i6,tonumber(itemID);
end

-- Print text
function DM:Chat(msg,r,g,b)
	if (DEFAULT_CHAT_FRAME) then
		if (not r and not g and not b) then r=1; g=1; b=1; end;
		if (not r) then r=0; end;
		if (not g) then g=0; end;
		if (not b) then b=0; end;
		DEFAULT_CHAT_FRAME:AddMessage(msg,r,g,b);
	end
end


-- Multi-threading utility
DM.MT={
	LastStack="",
	Current=0,
	Count=0,
	Speed=(1/30),
	LastTime=0,
	Threads={
	},
};

function DM.MT:Run(name,func,...)
	for _,tTable in pairs(self.Threads) do
		if (tTable.orig==func and tTable.name==name) then
			if (dbg) then dbg.print(name,"MT> Already running in another thread. Aborted."); end
			return;
		end
	end

	self.RunningCo=true;
	self.Count=self.Count+1;
--DM:Chat("Starting "..name);
	self.Threads[self.Count]={};
	self.Threads[self.Count].name=name;
	self.Threads[self.Count].orig=func;
	self.Threads[self.Count].cr=coroutine.create(func);
	self.LastTime=GetTime();
	self.LastStack="Running "..name;
	self.Speed=(1/30);
	-- For running without timer
	self.PassCounter=0;
	if (not self.Multiplier) then self.Multiplier=500; end

	local succeeded,result=coroutine.resume(self.Threads[self.Count].cr,...);
	if (not succeeded and Swatter) then
		if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
		else DM:Chat(result); DM:Chat(self.LastStack); end
	end
	self.RunningCo=nil;
end

--[[ Timer version
function DM.MT:Yield(immediate,dbdata)
	self.LastStack=debugstack(2);
	local now=GetTime();
	if (not immediate) then
		if (now-self.LastTime<self.Speed) then return; end
	end
	self.LastTime=now;		-- Inaccurate to account for other snags
	coroutine.yield();
	self.LastStack=debugstack(2);
end]]

-- Update version
function DM.MT:Yield(immediate,dbdata)
	self.LastStack=debugstack(2);
	self.PassCounter=self.PassCounter+1;
	if (not immediate) then
		if (self.PassCounter<self.Multiplier) then return; end
	end
	self.PassCounter=0;		-- Inaccurate to account for other snags
	coroutine.yield();
	self.LastStack=debugstack(2);
end

function DM.MT:Next(elapsed)
	if (self.Count<1) then return; end
	if (self.RunningCo) then return; end	-- Don't if we are already doing it. In case of real MT.
	-- Set timing for MT yield
	if (elapsed<self.Speed) then self.Multiplier=self.Multiplier+5;
	else self.Multiplier=self.Multiplier-6; end
	if (self.Multiplier<1) then self.Multiplier=1; end
--DM:Chat(self.Multiplier);

	if (not self.Threads[self.Current+1]) then self.Current=0; end	-- Wrap
	self.Current=self.Current+1;
	if (not self.Threads[self.Current]) then return; end	-- Nothing to do
	if (coroutine.status(self.Threads[self.Current].cr)=="dead") then
		local removeIt=self.Current;
--DM:Chat("Ending "..self.Threads[removeIt].name);
		while (self.Threads[removeIt]) do
			self.Threads[removeIt]=self.Threads[removeIt+1];
			removeIt=removeIt+1;
		end
		self.Current=self.Current-1;
		self.Count=self.Count-1;
	else
		self.RunningCo=true;
		local succeeded,result=coroutine.resume(self.Threads[self.Current].cr);
		if (not succeeded) then
			if (Swatter) then Swatter.OnError(result,nil,self.LastStack);
			else DM:Chat(result); DM:Chat(self.LastStack); end
		end
 		self.RunningCo=nil;
	end
end

function DM.MT:Processes()
	return self.Count;
end


-- Table compression utility
-- API:
-- DuckMod.Table:Init(who,UseCache,Base)
--	Set default options for your addon.
--	   who	  - An identifier for your add-on.
--	   UseCache - Use cached mode
--	   Base	 - Set a default table for read/write
--
-- DuckMod.Table:Read(who,entry,base)
--	Read a compressed entry
--	   who   - Your addon's ID, if any
--	   entry - The entry you want to read
--	   base  - The base table where 'entry' is located. If omitted,
--			   the default setting for your addon will be used.
--
-- DuckMod.Table:Write(who,entry,tData,base)
--	 Write an uncompressed entry
--	   who   - Your addon's ID, if any
--	   entry - The entry you want to write
--	   tData - Data write to 'entry'
--	   base  - The base table where 'entry' is located. If omitted,
--			   the default setting for your addon will be used.

-- v1
DM.Table={
	Scrap={},
	Default={
		Default={
			UseCache=true,
			Base=nil,
		},
	},
	CV=2,
	[1]={
		Entry ="\1",			--  prints as pipe after 4.3
			String="\1\1",
			Number="\1\2",
			Bool  ="\1\3",
			Nil   ="\1\4",
			Other ="\1\9",
		sTable="\2",			-- 
		eTable="\3",			-- 
		Version="\4",			-- 
			ThisVersion="1",
		Last="\4",
	},
	[2]={
		Entry ="\2",			-- 
			String="\2\7",
			Number="\2\2",
			Bool  ="\2\3",
			Nil   ="\2\4",
			Other ="\2\9",
		sTable="\3",			-- 
		eTable="\4",			-- 
		Version="\7",			--
			ThisVersion="2",
		Last="\7",
	},
	-- Same as last active
	Entry ="\2",			-- 
		String="\2\7",
		Number="\2\2",
		Bool  ="\2\3",
		Nil   ="\2\4",
		Other ="\2\9",
	sTable="\3",			-- 
	eTable="\4",			-- 
	Version="\7",
		ThisVersion="2",
	Last="\7",
};
-- Broken in WoW 4.3:
-- \1 \6


function DM.Table:Init(who,UseCache,Base)
	if (who=="Default") then return; end
	if (not Base) then return nil; end
	if (not self.Default[who]) then self.Default[who]={}; end
	wipe(self.Default[who]);
	self.Default[who].UseCache = UseCache;
	self.Default[who].Base	 = Base;
	self.Default[who].Cache	= {};
end

function DM.Table:GetType(typedata,tdExtra,CV)
	if (not CV) then CV=self.CV; end
	if (type(typedata)=="string") then return self[CV].String; end
	if (type(typedata)=="number") then return self[CV].Number; end
	if (type(typedata)=="boolean") then return self[CV].Bool; end
	if (type(typedata)=="table") then return self[CV].sTable..string.char(tdExtra); end
	if (not typedata) then return self[CV].Nil; end
	return self[CV].Other;
end

--tostring(123)  --Returns "123"
--tostring({})  --Returns "table: ###"
--tostring(function() end) --Returns "function: ###"
--tostring(nil) --Returns "nil"
--tostring(true) --Returns "true"

--[[
	MyTable = {
		First = {
			fdata=5,
		}
		second = {
			sdata="text",
		}
	}

	base[MyTable]="
	'Version''ThisVersion'sTable1'	-> Write
	'String'First'sTable2'			-> Compress (1)
	'String'fdata'Number'			-> Compress (2)
	5								-> Compress (3)
	'eTable2'						-> Compress (2)
	'String'second'sTable2'			-> Compress (1)
	'String'sdata'String'			-> Compress (2)
	text							-> Compress (3)
	'eTable2'						-> Compress (2)
	'eTable1'						-> Compress (1)
]]

function DM.Table:CompressV1(tData,level)
	local text="";
	if (type(tData)~="table") then return tostring(tData); end
	for entry,eData in pairs(tData) do
		text=text..self:GetType(entry,nil,1)..entry..self:GetType(eData,level,1)..self:CompressV1(eData,level+1);
		if (type(eData)=="table") then
			text=text..self[1].eTable..string.char(level);
		end
	end
	return text;
end

function DM.Table:CompressV2(tData,level)
	local text="";
	if (type(tData)~="table") then return tostring(tData); end
	for entry,eData in pairs(tData) do
		text=text..self:GetType(entry,nil,2)..entry..self:GetType(eData,level,2)..self:CompressV2(eData,level+1);
		if (type(eData)=="table") then
			text=text..self[2].eTable..string.char(level);
		end
	end
	return text;
end

function DM.Table:Write(who,entry,tData,base,section)
	local cache=self.Default[who].UseCache;
	if (not who) then who="Default"; end
	if (not base) then
		base=self.Default[who].Base;
	elseif (section and base~=self.Default[who].Base[section]) then		-- Only cache for base data
		cache=false
	end
	base[entry]=self[self.CV].Version..self[self.CV].ThisVersion..self:GetType(tData,100,self.CV)..self:CompressV2(tData,101);
	base[entry]=base[entry]..self[self.CV].eTable..string.char(100);

	if (cache) then
		if (base==self.Default[who].Base) then
			if (section) then
				if (not self.Default[who].Cache[section]) then self.Default[who].Cache[section]={}; end
				cache=self.Default[who].Cache[section];
			else
				cache=self.Default[who].Cache;
			end
			cache[entry]=tData;
		end
	end
	return true;
end

--[[
	index
	  1   2 -> Entry-type
	  3   a -> Entry name
	a+1 a+2 -> Data type
	a+3   b -> Data
]]

function DM.Table:PullEntry(cString,CV)
	if (not CV) then CV=self.CV; end
	local stop1,stop2,stop3;
	if (cString:sub(1,1)==self[CV].sTable) then
		stop1=cString:find(self[CV].eTable..cString:sub(2,2),3,true);
		if (not stop1) then return nil; end	-- Missing data
		return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1+2);
	end
	stop1=cString:find(self[CV].Entry,3,true); if (not stop1) then stop1=cString:len()+1; end	-- Virtual entry position
	stop2=cString:find(self[CV].sTable,3,true); if (not stop2) then stop2=cString:len()+1; end	-- Virtual entry position
	stop3=cString:find(self[CV].eTable,3,true); if (not stop3) then stop3=cString:len()+1; end	-- Virtual entry position
	if (stop1>stop2) then stop1=stop2; end
	if (stop1>stop3) then stop1=stop3; end
	return cString:sub(1,2),cString:sub(3,stop1-1),cString:sub(stop1);
end

function DM.Table:DecompressV1(base,cString,anon)
--DM:Chat("Decompressing...");
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,1);
			dData=self:DecompressV1(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,1);		-- Entry name
			if (eType==self[1].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,1);		-- Entry data
			if (dType==self[1].Number) then dData=tonumber(dData);
			elseif (dType==self[1].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[1].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[1].sTable) then
				base[eData]={};
				dData=self:DecompressV1(base[eData],dData);	-- dData is now without table-tags
			end
		end
		if (not base) then return dData;
		else
			if (anon) then base=dData;
			else base[eData]=dData; end
		end
	end
	return base;
end

function DM.Table:DecompressV2(base,cString,anon)
--DM:Chat("Decompressing...");
	while(cString:len()>0) do
		local eType,eData=nil,nil;
		local dType,dData=nil,nil;
		if (anon) then
			dType,dData,cString=self:PullEntry(cString,2);
			dData=self:DecompressV2(base,dData);	-- dData is now without table-tags
		else
			eType,eData,cString=self:PullEntry(cString,2);		-- Entry name
			if (eType==self[2].Number) then eData=tonumber(eData); end
			dType,dData,cString=self:PullEntry(cString,2);		-- Entry data
			if (dType==self[2].Number) then dData=tonumber(dData);
			elseif (dType==self[2].Bool) then if (dData=="true") then dData=true; else dData=false; end
			elseif (dType==self[2].Nil) then dData=nil;
			elseif (dType:sub(1,1)==self[2].sTable) then
				base[eData]={};
				dData=self:DecompressV2(base[eData],dData);	-- dData is now without table-tags
			end
		end
		if (not base) then return dData;
		else
			if (anon) then base=dData;
			else base[eData]=dData; end
		end
	end
	return base;
end

-- "\1" 
-- "\2" 
-- "\3" 
-- "\4" 

--	Entry ="\1", 
--	sTable="\2", 
--	eTable="\3", 


function DM.Table:Read(who,entry,base,section)
	local cache=self.Default[who].UseCache;
	if (not who) then who="Default"; end								-- Using default settings
	if (not base) then
		base=self.Default[who].Base;
	elseif (section and base~=self.Default[who].Base[section]) then		-- Only cache for base data
		cache=false
	end
	if (not base[entry]) then return nil; end							-- Entry does not exist
	if (type(base[entry])~="string") then			-- Not a proper entry
		DM:Chat("Got "..type(base[entry]).." as entry");
		return nil;
	end
--	if (base[entry]:byte(1)~=self.Version) then return base[entry]; end	-- It's uncompressed data
	local CV=tonumber(base[entry]:sub(2,2));
	local fnDecompress=self["DecompressV"..CV];		-- Look for decompressor
	if (not fnDecompress) then return nil; end		-- We don't have a decompressor for this data
	-- Not using cache, so decompress to scrap-book and return it
	if (not cache) then
		wipe(self.Scrap);
		return DM:CopyTable(fnDecompress(self,self.Scrap,base[entry]:sub(3),true));		-- Strip version
	end
	if (section) then
		if (not self.Default[who].Cache[section]) then self.Default[who].Cache[section]={}; end
		cache=self.Default[who].Cache[section];
	else
		cache=self.Default[who].Cache;
	end
	-- Not decompressed yet, so decompress to cache and return it
	if (not cache[entry]) then
--		cache[entry]=DM:CopyTable(fnDecompress(self,self.Scrap,base[entry]:sub(3),true));
		cache[entry]={};
		fnDecompress(self,cache[entry],base[entry]:sub(3),true);
	end
	return cache[entry];
end

function DM.Table:PurgeCache(who)
	wipe(self.Default[who].Cache);
	collectgarbage("collect");
end


DM.Code={};

--[[
	Code with password. It's not complicated, but needs the password
	to decode, and is totally unreadable after coding.
		code=mod(byte+seed+pw[next],256);
		byte=mod(code-pw[next]-seed,256);
]]
function DM.Code:Protect(text,pw)
	if (not pw or pw=="") then return text; end
	local seed=0;
	local pwlen=pw:len();
	for i=1,pwlen do seed=seed+pw:byte(i); end
	local code="";
	local pi=1;
	for i=1,text:len() do
		code=code..string.char(mod(text:byte(i)+(seed+pw:byte(pi)),256));
		pi=pi+1; if (pi>pwlen) then pi=1; end
	end
	return self:Code(code);		-- Results in pure 8-bit, so we must code it
end

function DM.Code:Unprotect(code,pw)
	if (not pw or pw=="") then return code; end
	local decoded;
	code,decoded=self:Decode(code);
	if (decoded~=true) then return code; end
	local seed=0;
	local pwlen=pw:len();
	for i=1,pwlen do seed=seed+pw:byte(i); end
	local text="";
	local pi=1;
	for i=1,code:len() do
		local value=mod(code:byte(i)-(seed+pw:byte(pi)),256);
		while(value<0) do value=value+256; end
		text=text..string.char(value);
		pi=pi+1; if (pi>pwlen) then pi=1; end
	end
	return text;
end

--[[
	Text-code any three 8-bit values to four values ranging from 64-127
	inclusive.
		composite=byte1*65536+byte2*256+byte3
		char1=mod(composite,64)+64;
		composite=composite/64;
		char2=mod(composite,64)+64;
		composite=composite/64;
		char3=mod(composite,64)+64;
		char4=composite/64+64;

		composite=(char4-64)*262144+(char3-64)*4096+(char2-64)*64+(char1-64);
		char3=mod(composite,256);
		composite=composite/256;
		char2=mod(composite,256);
		char1=composite/256;
]]

--	raw string bytes:
--	11111111 22222222 33333333

--	Binary coding: +64 >> 7
--	111111110000000000000000
--	000000002222222200000000
--	000000000000000033333333

-->	111111112222222233333333 -> com[0]=0333333d
--	000000111111112222222233 -> com[1]=0222233c
--	000000000000111111112222 -> com[2]=0112222b
--	000000000000000000111111 -> com[3]=0111111a
function DM.Code:Code(text)
	text=text:len()..":"..text;
	while (math.floor(text:len()/3)*3~=text:len()) do text=text.." "; end
	local code="";
	for i=1,text:len(),3 do
		local com=bit.lshift(text:byte(i),16)+bit.lshift(text:byte(i+1),8)+text:byte(i+2);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(mod(com,64)+64); com=bit.rshift(com,6);
		code=code..string.char(com+64);
	end
	return code;
end

--	raw string bytes:
--	0333333d 0222233c 0112222b 0111111a

--	Binary coding:
--	111111000000000000000000
--	000000112222000000000000
--	000000000000222233000000
--	000000000000000000333333
-->	111111112222222233333333 -> inp=33333333
--	000000001111111122222222 -> inp=22222222 33333333
--	000000000000000011111111 -> inp=11111111 22222222 33333333

-- Return: text (string), decoded (true|false)
function DM.Code:Decode(code)
	local text="";
	if (mod(code:len(),4)~=0) then return code,false; end
	for i=1,code:len(),4 do
		local com;
		com=bit.lshift(code:byte(i+3)-64,18);
		com=com+bit.lshift(code:byte(i+2)-64,12);
		com=com+bit.lshift(code:byte(i+1)-64,6);
		com=com+(code:byte(i)-64);
		local inp=string.char(mod(com,256)); com=bit.rshift(com,8);
		inp=string.char(mod(com,256))..inp; com=bit.rshift(com,8);
		inp=string.char(mod(com,256))..inp;
--		inp=string.char(com)..inp;
		text=text..inp;
	end
	local here=text:find(":");
	if (not here) then return code,false; end
	local size=text:sub(1,here-1); size=tonumber(size);
	if (not size or size<1) then return code,false; end
	text=text:sub(here+1);
	size=text:len()-size;
	if (size>3 or size<0) then return "",true; end
	if (size>0) then text=text:sub(1,text:len()-size); end
	return text,true;
end

-- WoWnet
DM.WoWnet={
	prefix="Wnet",
	Connected=nil,				-- Not connected yet
	Session={					-- Data and functions specific for sessions
		LoopBack=nil,			-- Running in loopback mode. i.e playing the madman (talk to self)
		LastPubOut=nil,			-- Last data sent to public channel
		cbSearchResult=nil,		-- Callback: Incoming search-results
		ResolveWait=nil,		-- Stuff waiting WoWnet resolving
	},
	LinkColour="ffff8080";
};

-- The host-data for all servers on this ream/faction.
WoWnetServer={			-- Complete server contents (this server)
};

--[[
	host
		section
			SECTIONDATA
				Keys="key1,key2,...,keyN"
				Data=page-data
				Description="Short description"
			subsection
				...
----------------------------------------
	contents
		host1
			section 1
				section 1/1
					section 1/1/1
					section 1/1/2
				section 1/n
			section 2
				section 2/1
		host2
			...
]]

function DM.WoWnet:Link(location,text)
	return "|c"..DM.WoWnet.LinkColour.."|Hwownet:"..location.."|h["..text.."]|h|r";
end


-- contents -> Table for server contents
function DM.WoWnet:Connect(cbSR,cbID)
	local channel="WoWnet"
--	if (not JoinChannelByName(channel)) then
	if (not JoinTemporaryChannel(channel)) then
		local count=1
		self.Connected=nil;
		DM:Chat("Someone has blocked the WoWnet channel on this server.");
		while(not JoinTemporaryChannel("WoWnet"..count)) do
			count=count+1;
		end
		channel="WoWnet"..count;
		DM:Chat("Open WoWnet found at "..channel);
	end
	self.Connected=GetChannelName(channel);
	DM:Chat("Connected to WoWnet - "..type(self.Connected).." - "..self.Connected);
	if (self.Connected) then
		self.Session.cbSearchResult=cbSR;
		self.Session.cbInData=cbID;
		RegisterAddonMessagePrefix(DM.WoWnet.prefix);
--		DM.Table:Init(DM.WoWnet.prefix,nil,WoWnetServer);
	else
		self.Session.cbSearchResult=nil;
		self.Session.cbInData=nil;
	end
	return self.Connected;
end

-- Traverse all keys for this section and collate matched locations
function DM.WoWnet:CheckKeys(section,sTable,key)
	local list={};
	for entry,eTable in pairs(sTable) do
		if (entry~="SECTIONDATA") then
			list=DM:CopyTable(self:CheckKeys(section.."/"..entry,eTable,key),list);
		else
			if (eTable.Keys) then
				local param={strsplit(",",eTable.Keys)};
				for _,tag in pairs(param) do
					tag=string.lower(strtrim(tag));
					if (tag==key) then
						if (not eTable.Description) then list[section]="<no description>";
						else list[section]=eTable.Description; end
						break;
					end
				end
			end
		end
	end
	return list;
end

-- Traverse all hosts for the key
function DM.WoWnet:FindKey(key)
	key=string.lower(strtrim(key));
	local list={};
	if (WoWnetServer[GetRealmName()]) then
		for host,hTable in pairs(WoWnetServer[GetRealmName()]) do
			list[host]={};
			for section,sTable in pairs(hTable) do
				list[host]=DM:CopyTable(self:CheckKeys("/"..section,sTable,key),list[host]);
			end
		end
	end
	return list;
end

-- This function receives all information from the WoWnet channel
-- sender -> Data originator
-- text -> The data
function DM.WoWnet:Input(sender,text)
	-- A search-phrase received
	if (text:find("search:",1,true)==1) then
		text=string.sub(text,8);
		-- Get section-list for supplied keys
		local param={strsplit(",",text)};
		local list={};
		local index=1;
		while (param[index]) do						-- Do all supplied keys
			local found=self:FindKey(param[index]);
			list=DM:CopyTable(found,list);				-- Overlay findings
			index=index+1;
		end
		-- Check the list
		local found=nil;
		for host,hTable in pairs(list) do
			found=true;
			break;
		end
		if (not found) then return; end				-- No matches at all
--TestDebugTable=DM:CopyTable(list);
		-- There are matches, so "answer the general, Baldrick"
		for host,hTable in pairs(list) do
			for section,desc in pairs(hTable) do
				self.Session:SendContents(sender,host,section..DUCKNET_WNSPLIT1..desc);
			end
		end
	elseif (text:find("goloc:",1,true)==1) then
--		DM:Chat("Location request: "..text:sub(7),1);
		self.Session:HandleAddress(sender,text:sub(7));
	elseif (text:find("ping:",1,true)==1) then
		DM:Chat("Ping request: "..text:sub(6),1);
		if (WoWnetServer[GetRealmName()][text:sub(6)]) then
			DM.WoWnet.Session:SendData(sender,"ping",{Request=text});
		end
	end
end

function DM.WoWnet.Session:HandleAddress(sender,text)
	local splitter=text:find("/",1,true);
	local who,loc;
	if (splitter) then
		who=text:sub(1,splitter-1);
		loc=text:sub(splitter+1);
	else
		who=text;
		loc=""
	end
--	DM:Chat("Testing who: "..who,1);
	if (not WoWnetServer[GetRealmName()][who]) then return; end
--	DM:Chat(who.." is here.",1);
	if (loc=="") then
		loc="home";
		if (not WoWnetServer[GetRealmName()][who][loc]) then
			loc="default";
			if (not WoWnetServer[GetRealmName()][who][loc]) then
				loc="index";
				if (not WoWnetServer[GetRealmName()][who][loc]) then
					return;
				end
			end
		end
	end
	if (not WoWnetServer[GetRealmName()][who][loc]["SECTIONDATA"]) then
		return;
	end
	local theData=DM:CopyTable(WoWnetServer[GetRealmName()][who][loc]["SECTIONDATA"]);
	theData.Meta={
		Server=who,
		Location=text,		-- As it would be specified in the edit-box
	};
	self:SendData(sender,text,theData);
end

function DM.WoWnet.Session:Search(data)
	self:SendPublic("search:"..data);
end

function DM.WoWnet.Session:GoTo(data)
	DM.Net:ConnectW(DM.WoWnet.prefix,DM.WoWnet.Session.cbInData,"");
	self:SendPublic("goloc:"..data);
end

function DM.WoWnet.Session:SendPublic(data)
	DM.WoWnet.Session.LastPubOut=data;
	SendChatMessage(data,"CHANNEL",nil,DM.WoWnet.Connected);
end

function DM.WoWnet.Session:HandleLink(link)
	local lType,lAddress=strsplit(":",link);
	if (lType=="wownet") then
--		DM:Chat("Clicked "..lType.." link: "..lAddress,1);
		DM.WoWnet.Session:GoTo(lAddress);
	end
end

function DM.WoWnet.Session:InDataIntercept(prefix,user,sender,data)
	if (not user) then return nil; end
	-- Decode to a table, as outer table has not been coded
	for entry,_ in pairs(data) do
		data[entry]=DM.Table:GetType({},100)..DM.Code:Decode(data[entry])..DM.Table.eTable..string.char(100);	-- Decode
		data[entry]=DM.Table:DecompressV1({},data[entry],true);	-- Unpack
	end

	-- Check special-handling data
	if (data.ping) then
		DM:Chat("user: "..user);
		DM:Chat("sender: "..sender);
		if (data.ping and data.ping.Request) then
			if (data.ping.Request:find("ping:")==1) then
				if (data.ping.Request:sub(6)==DM.WoWnet.Session.ResolveWait.Server) then
					DM:Chat("request: "..data.ping.Request);
					DM.WoWnet.Session.ResolveWait.PipeToon=sender;	-- Save alt
				end
			end
		end
		return true;
	end

	self.cbInData(data,sender);	-- Give table-pointer to registered WoWnet addon
	return true;
end


-- receiver -> The real user to send the message to
-- host -> The host to request a session with ("user" may be an alt of the host)
-- data -> Data to send
function DM.WoWnet.Session:SendContents(receiver,host,data)
--	DM:Chat("-> WN user: "..receiver,1);
--	DM:Chat("-> WN host: "..host,1);
--	DM:Chat("-> WN data: "..data,1);
	if ((string.len(host)+3+string.len(data)+1)>255) then return nil; end
	DM.WoWnet.Session:SAM("wNc"..host..":"..data,"WHISPER",receiver);
	return true;
end

-- receiver -> Who to /w to
-- location -> The address of the data following
-- data - The data for this location
function DM.WoWnet.Session:SendData(receiver,location,data)
--DM:Chat("ConnectW to "..receiver,1);
	DM.Net:ConnectW(DM.WoWnet.prefix,DM.WoWnet.Session.cbInData,receiver);
--DM:Chat("Sending table for "..location,1);

	local prefix=DM.WoWnet.prefix;
	if (not DM.Net:CanTransmit(prefix)) then return false; end					-- Can't transmit now
	DM.Net:ClearOutput(prefix);
	DM.Net:AddKey(prefix,"S",0);
	DM.Net:AddKey(prefix,"M","SendData");
	DM.Net:AddKey(prefix,"A",DUCKNET_ACT_TRANSMIT);
	DM.Net:NewLine(prefix);
	local text;
	text=DM.Code:Code(DM.Table:CompressV1(data,101));
	if (not DM.Net:AddEntry(prefix,location,text)) then return nil; end
	return DM.Net:DoTransmission(prefix);
end

-- SAM: Send Addon Message
function DM.WoWnet.Session:SAM(a1,a2,a3)
	SendAddonMessage(DM.WoWnet.prefix,a1,a2,a3);
end

function DM.WoWnet.Session:SendError(prefix,header,message)
end

function DM.WoWnet.Session:Error(header,message)
	DM:Chat(header..": "..message,1);
end

-- wNr -> Request
-- wNc -> Contents
-- wNd -> Data

-- Called when a "WHISPER" with "wNc" is received, denoting incoming contents-entry
-- sender -> Player to send info to
-- host -> Host for which contents is requested
-- data -> section DUCKNET_WNSPLIT1 description
function DM.WoWnet.Session:Contents(sender,host,data)
--DM:Chat("Ready...",1);
	if (not self.cbSearchResult) then return; end
--DM:Chat("Cleared...",1);
	self.cbSearchResult(sender,host,strsplit(DUCKNET_WNSPLIT1,data));
end

-- A table
function DM.WoWnet.Session.INFO(input)
	local info=DM:CopyTable(input);

	local marker=nil;

	local seq=1;
	while (info[seq]) do
		if (info[seq].Type=="M") then marker=info[seq].Command; end
		seq=seq+1;
	end

	DM.WoWnet.Session.InMarker=marker;
	if (marker) then
		if (marker=="PageData") then
		end
	end
end


-- DuckNet
DM.Net={
	Split1=DUCKNET_FREESPLIT1,
	Split2=DUCKNET_TREESEP,
	Timers = {
		Last=0,
		LastEntry=0,
	},
	DB = { },
	TheFrame=nil,
};


function DM.Net:NegWindow()
	return DUCKNET_NEG_DONE;
end

-- Set up for handling
function DM.Net:Init()
	if (not self.TheFrame) then
		self.TheFrame=CreateFrame("Frame","DuckMod-Net-Messager"..TV,UIParent);
		if (not self.TheFrame) then
			DM:Chat("Could not create a DuckMod frame",1);
			return;
		end
	end
	if (not self.TheFrameWorldMap) then
		self.TheFrameWorldMap=CreateFrame("Frame","DuckMod-Net-Messager"..TV,WorldMapDetailFrame);
		if (not self.TheFrameWorldMap) then
			DM:Chat("Could not create a DuckMod frame (2)",1);
			return;
		end
	end
	self.TheFrame:RegisterEvent("CHAT_MSG_ADDON");				-- Addon channel
	self.TheFrame:RegisterEvent("CHAT_MSG_CHANNEL");			-- WoWnet
	self.TheFrame:RegisterEvent("CHAT_MSG_CHANNEL_NOTICE");		-- WoWnet
	self.TheFrame:RegisterEvent("GUILD_ROSTER_UPDATE");			-- Database
	self.TheFrame:RegisterEvent("WHO_LIST_UPDATE");				-- Database
	self.TheFrame:SetScript("OnEvent",DM.Net.OnEvent);
	self.TheFrame:SetScript("OnUpdate",DM.Net.HeartBeat);
	self.TheFrameWorldMap:SetScript("OnUpdate",DM.Net.HeartBeat);


	-- As LUA random can be dodgy at times, it is recommended to make
	-- a few calls to it to "get it going"
	math.random(); math.random(); math.random();
end


function DM.Net:Valid(prefix,ctype)
	if (not prefix) then return nil; end							-- None provided
	if (prefix==DM.WoWnet.prefix and DM.WoWnet.Connected and ctype=="WHISPER") then
		return true;
	end
	if (not self.DB[prefix]) then return nil; end				-- Provided, but not registered
	if (ctype) then
		if (self.DB[prefix].CType~=ctype) then return nil; end	-- Wrong channel type
	end
	return true;
end


-- The heartbeat. The OnUpdate handler
function DM.Net.HeartBeat(frame,elapsed)
--	local elapsed=arg1;
	if (not elapsed) then return; end
	if (not dbg and DuckModDebug) then dbg=DuckModDebug; end

	DM.MT:Next(elapsed);		-- Run multi-threading

	-- Second pass of rendering
	if (DM.RenderWnCounter==2) then DM:Render_2(DM.RU.Canvas,DM.RenderWnRemaining); end

	if (DM.Database.BootDelay) then
		DM.Database.BootDelay=DM.Database.BootDelay-elapsed;
		if (DM.Database.BootDelay<=0) then DM.Database.BootDelay=nil; end
		return;
	end

	DM.Net.Timers.Last=DM.Net.Timers.Last+elapsed;					-- Update this
	if (DM.Net.Timers.Last<DUCKNET_HEARTBEAT) then return; end		-- Heartbeat is DUCKNET_HEARTBEAT seconds
	DM.Net.Timers.Last=0;											-- Restart

	-- Maintain offline user supressor list
	if (next(DM.SentBlindWhisperListOfReceivers)) then
		local toon,val=next(DM.SentBlindWhisperListOfReceivers);
		if (time()-val>5) then DM.SentBlindWhisperListOfReceivers[toon]=nil; end
	end

	-- WoWnet data waiting for resolving
	if (DM.WoWnet.Session.ResolveWait) then
		if (DM.WoWnet.Session.ResolveWait.Table and DM.WoWnet.Session.ResolveWait.Server and DM.WoWnet.Session.ResolveWait.PipeToon) then
DM:Chat("ready to send formdata");
			if (DM.WoWnet.Session:SendData(DM.WoWnet.Session.ResolveWait.PipeToon,"formdata:"..DM.WoWnet.Session.ResolveWait.Server,DM.WoWnet.Session.ResolveWait.Table)) then
				wipe(DM.WoWnet.Session.ResolveWait); DM.WoWnet.Session.ResolveWait=nil;	-- Clear if ok
			end
		end
	end

	local now=time();
	for k,v in pairs(DM.Net.DB) do
		if (v.Alive) then										-- An alive-stamp exists (wownet)
			if (now>=v.Alive) then								-- outdated pipe
				if (dbg) then dbg.print(k,DM.Color.Yellow.."Idle private channel "..k.." has been closed|r"); end
				DM.Script:SetValue(k,"PIPERCONNECTION",0);		-- kill the piper connection if any
				DM.Net:Disconnect(k);							-- Delete this entry
			else DM.Net:HeartBeatCycle(k); end
		else
			DM.Net:HeartBeatCycle(k);								-- non-wownet
		end
	end

	-- Note that we're not using the accurate form. The reason is that this code
	-- does not need accurate timing, and this way has a much bigger chance of
	-- success if any hick-ups should occur elsewhere - thus increasing the chance
	-- of correct transmission.
end


-- Handler for a single prefix
function DM.Net:HeartBeatCycle(prefix)
	if (not self:Valid(prefix)) then return 1; end		-- Invalid prefix

	local now=time();
	DM.Database:Heartbeat(prefix,now);

--[[		DuckNet negotiation		]]
	if (self.DB[prefix].Negotiate.Start) then
		self.DB[prefix].Negotiate.HoldOff=DUCKNET_NEG_RUNNING;
		self.DB[prefix].Negotiate.Start=false;
		return;
	end
	if (self.DB[prefix].Negotiate.HoldOff>=DUCKNET_NEG_RUNNING) then
		self.DB[prefix].Negotiate.HoldOff=self.DB[prefix].Negotiate.HoldOff+DUCKNET_HEARTBEAT;
		if (self.DB[prefix].Negotiate.HoldOff>=DUCKNET_NEG_DONE) then
			if (self.DB[prefix].Negotiate.StartMarker) then
				local IWon=false;
				if (self.DB[prefix].Negotiate.SelfStamp>=self.DB[prefix].Negotiate.HighestStamp) then				-- I have highest stamp
					if (self.DB[prefix].Negotiate.HighestRoll<self.DB[prefix].Negotiate.Roll) then IWon=true; end;	-- I have highest roll
				end
				if (dbg) then dbg.DN_NegotiationComplete(prefix,IWon,self.DB[prefix].Negotiate); end
				if (self.DB[prefix].Negotiate.StartMarker:match("^DMDBLIB.+")) then
					if (DM.Database.Addon[prefix].PollingKey) then
						if (dbg) then
							dbg.print(prefix,DM.Color.Green..DM.Database.Util.Debug:GetLoc(DM.DatabaseQueue[prefix][DM.Database.Addon[prefix].PollingKey]).." has been removed from the sync-queue");
							if (not next(DM.DatabaseQueue[prefix])) then dbg.print(prefix,DM.Color.Green.."Sync-queue is now empty"); end
						end
						DM.Database.Addon[prefix].PollingKey=nil;
						DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;
					end
					DM.Database.Util:NegotiateDone(prefix,self.DB[prefix].Negotiate.StartMarker,IWon);
				elseif (self.DB[prefix].CallBack.NegotiateWon) then		-- Check for callback-function
					self.DB[prefix].CallBack.NegotiateWon(IWon,self.DB[prefix].Negotiate.StartMarker);		-- Tell the calling addon if it won or not
				end
			end
			self:StopNegotiation(prefix);
		end
	end

	-- Handle transmission
	if (self.DB[prefix].Transmitting>DUCKNET_TRANS_STOP) then	-- It's transmitting
		if (self.DB[prefix].TransmitDelay>0) then
			self.DB[prefix].TransmitDelay=self.DB[prefix].TransmitDelay-DUCKNET_HEARTBEAT;
		elseif (not self.DB[prefix].LastOutput) then					-- It's there now
			self:SendNextLine(prefix);							-- Duh
		end
	end

	-- Check broken incoming data
	if (self.DB[prefix].LastEntry>-1 and now-self.Timers.LastEntry>10) then
		self.DB[prefix].LastEntry=-1;								-- Stop receiving
		self.DB[prefix].Rebuilding=nil;
	end

	if (self.DB[prefix].Common.Data and self.DB[prefix].Common.Boot>=0) then
		self.DB[prefix].Common.Boot=self.DB[prefix].Common.Boot-DUCKNET_HEARTBEAT;
		if (self.DB[prefix].Common.Boot<0) then
			-- Process boot
		end
	end

	return 0;
end

-- An event has been received
function DM.Net:OnEvent(event,arg1,arg2,arg3,arg4,_,_,_,_,arg9)
	if (event=="CHAT_MSG_ADDON") then
		if (DM.SentBlindWhisperListOfReceivers[arg4]) then DM.SentBlindWhisperListOfReceivers[arg4]=nil; end
		-- Check for WoWnet input
		if (arg1==DM.WoWnet.prefix and arg3=="WHISPER") then
			if (arg2:find("wN")==1) then
--DM:Chat("<- "..arg2,1);
--DM:Chat("<- "..arg3,1);
--DM:Chat("<- "..arg4,1);
				local splitter=arg2:find(":");
				local host=arg2:sub(1,splitter-1);
				arg2=arg2:sub(splitter+1);
--DM:Chat("<- "..splitter,1);
--DM:Chat("<- "..host,1);
--DM:Chat("<- "..arg2,1);

				host=host:sub(3);
				if (host:find("r")==1) then			-- Request received
					host=host:sub(2);
					if (DM.WoWnet.Session:Request(arg4,host,arg2)) then return; end
				elseif (host:find("c")==1) then		-- Contents received
					host=host:sub(2);
-- DM:Chat("<- "..arg1,1); DM:Chat("<- "..arg2,1); DM:Chat("<- "..arg3,1); DM:Chat("<- "..arg4,1);
					if (DM.WoWnet.Session:Contents(arg4,host,arg2)) then return; end
				end
			end
		end
--arg1	prefix
--arg2	message
--arg3	distribution type ("PARTY", "RAID", "GUILD", "BATTLEGROUND" or "WHISPER")
--arg4	sender

		-- Normal input
		if (dbg) then dbg.DN_CHAT_MSG_ADDON(arg1,arg2,arg3,arg4); end
-- printing arg2,arg3,arg4 here
		if (not DM.Net:Valid(arg1,arg3)) then return; end;	-- Quick way out without further function-calls - which ParseInput will do
		DM.Net:ParseInput(arg1,arg2,arg4);					-- It's from a registered channel, so attempt to decode it
		return;
	end
	if (event=="CHAT_MSG_CHANNEL" and strlower(arg9:sub(1,6))=="wownet") then
		if (arg1~=DM.WoWnet.Session.LastPubOut) then
			DM.WoWnet:Input(arg2,arg1);
		else
			DM.WoWnet.Session.LastPubOut=nil;
		end
		return;
--arg1	chat message
--arg2	author
--arg8	channel number
--arg9	channel name without number (this is _sometimes_ in lowercase)
	end
	if (event=="CHAT_MSG_CHANNEL_NOTICE" and arg9:lower()=="wownet") then
		-- reconnect if kicked
		return;
--arg1	"YOU_JOINED", "YOU_LEFT", or "THROTTLED"
--arg9	channel name without number (this is _sometimes_ in lowercase)
	end

	-- Database inter-guild local roster update
	if (event=="GUILD_ROSTER_UPDATE") then
print("GUILD_ROSTER_UPDATE");
		if (arg1) then return; end		-- Some change, so don't all start sync together
		local list={};
		local num=GetNumGuildMembers();
		for i=1,num do
			local name=GetGuildRosterInfo(i);
			list[name]=DM.Database.PIPER_IDLE;
		end
		local guild=GetGuildInfo("player");
		if (not guild) then return; end
print(num.." members in "..guild);
		DM.Database.Info:UpdateGuildMemberList(guild,list);
	end

	-- Database inter-guild external roster update
	if (event=="WHO_LIST_UPDATE") then
print("WHO_LIST_UPDATE");
		local list={};
		local num=GetNumWhoResults();
		local guild;
		for i=1,num do
			_,guild=GetWhoInfo(i);
			if (not list[guild]) then list[guild]=1;
			else list[guild]=list[guild]+1; end
		end
		local goal,amount="",0;
		for guild,count in pairs(list) do
			if (count>amount) then goal=guild; amount=count; end
		end
		guild=goal;
		wipe(list);
		for i=1,num do
			local name,gname=GetWhoInfo(i);
			if (gname==guild) then list[name]=DM.Database.PIPER_IDLE; end
		end
		DM.Database.Info:UpdateGuildMemberList(guild,list,true);	-- Merge it with the other list, if any
	end
end
--
--	name,guild,level,race,class,zone,classFileName,sex=GetWhoInfo(index)


-- Handles basic set-up for network negotiation
function DM.Net:StartNegotiation(prefix,stampmarker,selfstamp,linestamp)
	if (dbg) then dbg.DN_Negotiation_Start(prefix,"self",stampmarker,selfstamp,linestamp); end
	self.DB[prefix].Negotiate.StartMarker=stampmarker;
	self.DB[prefix].Negotiate.SelfStamp=selfstamp;
	if (not linestamp) then linestamp=selfstamp; end;
	self.DB[prefix].Negotiate.Start=true;											-- Start a negotiation
	self.DB[prefix].Negotiate.Roll=0;												-- Do not participate in roll yet
	self.DB[prefix].Negotiate.HighestStamp=0;
	self.DB[prefix].Negotiate.HighestRoll=0;

	if (selfstamp<=linestamp) then														--> I have same or lower
		self.DB[prefix].Negotiate.HighestStamp=linestamp;							-- I am not higher, so set incoming as highest
		if (dbg) then dbg.Log(prefix,"NegStart -> I have same or lower stamp (my poll or incoming)"); end
		return nil;			-- Did not send data
	end

	self.DB[prefix].Negotiate.HighestStamp=selfstamp;							-- I am higher, so set it as highest
	self.DB[prefix].Negotiate.Roll=math.random(1000000);							-- Make my own random for transmission
	local output="";																-- Start line for transmission
	if (stampmarker) then output=output..self:MakeEntry("M",stampmarker); end	-- Add stampmarker if it has been received
	output=output..self:MakeEntry("S",self.DB[prefix].Negotiate.SelfStamp);	-- Add own stamp
	output=output..self:MakeEntry("R",self.DB[prefix].Negotiate.Roll);		-- Add own random number
	output=output..self:MakeEntry("A",DUCKNET_ACT_MYSTAMP);						-- Set action
	self:Out(prefix,output);														-- Send your random
	if (dbg) then dbg.Log(prefix,"NegStart -> Roll (out): "..self.DB[prefix].Negotiate.Roll); end
	return true;				-- Data sent
end


-- Kill it if need be
function DM.Net:StopNegotiation(prefix)
	self.DB[prefix].Negotiate.HighestStamp=0;
	self.DB[prefix].Negotiate.HighestRoll=0;
	self.DB[prefix].Negotiate.Start=false;								-- Don't start a negotiation
	self.DB[prefix].Negotiate.HoldOff=DUCKNET_NEG_STOPPED;				-- No negotiation is running
end


--[[ DuckNet Protocol 1 ]]
function DM.Net:ParseInput(prefix,text,sender)
	local function Closest()
		local small=text:find(DUCKNET_DATA); if (not small) then small=10000; end
		local Colon=text:find(DUCKNET_COMMAND); if (not Colon) then Colon=10000; end
		local Hash=text:find(DUCKNET_ENTRY); if (not Hash) then Hash=10000; end
		if (Colon<small) then small=Colon; end
		if (Hash<small) then small=Hash; end
		return small;
	end
	local function Next()
		local code=nil;
		local info=nil;
		local data=nil;
--[[		Commands			]]
		if (text:sub(1,1)==DUCKNET_COMMAND) then
			code=text:sub(1,2); text=text:sub(3);		-- Extract command
			local near=Closest();
			if (near>1) then info=text:sub(1,near-1); text=text:sub(near);		-- Extract the info
			else info=text; text=""; end														-- Use the rest of the line as info
--[[		Table entry			]]
		elseif (text:sub(1,1)==DUCKNET_ENTRY) then
			code=text:sub(1,2); text=text:sub(3);									-- Extract tag and type-indicator
			local near=Closest(); if (near>1) then												-- Locate next delimiter
				info=text:sub(1,near-1); text=text:sub(near);						-- Extract entry name
				if (text:sub(1,1)==DUCKNET_DATA) then
					text=text:sub(2);													-- Remove data-marker if data is present
					if (text:sub(1,1)=="\"") then
						text=text:sub(2); near=text:find("\"");					-- Remove first quote. Find next quote
						if (not near) then text="\""..text; data=text; text="";					-- Not there. Reinsert previous. Copy. Save.
						else data=text:sub(1,near-1); text=text:sub(near+1);		-- Get it and remove including next quote
						end
					else
						near=Closest();				-- Not string, get next delimiter
						if (near==1) then data=nil;
						elseif (near>1) then data=tonumber(text:sub(1,near-1)); text=text:sub(near);
						else data=tonumber(text); text=""; end
					end
				end
			else info=text; text=""; end		-- Use all
		else info=text; text=""; end			-- Use all
		return code,info,data;
	end
---------- CODE START ----------
	if (not self:Valid(prefix)) then return; end		-- Validate incoming prefix

	-- InStamp also at own stuff
	local now=time();
	DM.Talk[prefix]=now;
	if (self.DB[prefix].CallBack.InStamp) then self.DB[prefix].CallBack.InStamp(now); end

	-- Base evaluation
	if (text==self.DB[prefix].LastOutput) then self.DB[prefix].LastOutput=nil; return; end	-- Own transmission received
	if (not text) then return; end
	text=self:SwapText(text,DUCKNET_TOCODE,DUCKNET_FROMCODE);

	-- Check for broken lines, for concatenation
	local broken=nil;
	if (text:find(DUCKNET_ENTRYBREAKCONTINUE)==text:len()-(DUCKNET_ENTRYBREAKCONTINUE:len()-1)) then	-- Break-mark
		broken=text:sub(1,text:len()-DUCKNET_ENTRYBREAKCONTINUE:len());
	end
	if (broken and not self.DB[prefix].Rebuilding) then self.DB[prefix].Rebuilding=""; end	-- Start rebuilding
	if (self.DB[prefix].Rebuilding) then											-- We are concatenating something
		if (broken) then												-- There's more coming
			self.Timers.LastEntry=now;
			self.DB[prefix].Rebuilding=self.DB[prefix].Rebuilding..broken;
			return;
		end
		text=self.DB[prefix].Rebuilding..text;										-- We're done now
		self.DB[prefix].Rebuilding=nil;											-- Go back to normal operation
	end

	-- Set-up for decoding
	local linestamp=0;
	local tp=self.DB[prefix].Data;					-- Pointer to the table root
	self.DB[prefix].Info={ };						-- Clear the info-table
	local infoline=1;									-- To build the info table
	local stampmarker=nil;								-- Stamp marker for this line
	local lastroll=0;
	if (not self.DB[prefix].DBInfo) then self.DB[prefix].DBInfo={}; end

--[[	Cycle through the entire line of input	]]
	while(text:len()>0) do
		local tag,entry,data=Next();					-- Extract next entry
		if (not tag) then break; end;					-- Some tag-less info
		local entryn=tonumber(entry);					-- Better execution

		--[[					]]
		--[[   Common entries   ]]
		--[[					]]

		--[[ Stamp ]]
		if (tag==DUCKNET_COMMAND.."S") then
			linestamp=entryn;

		--[[ Random-data ]]
		elseif (tag==DUCKNET_COMMAND.."R") then
			lastroll=entryn;
			if (self.DB[prefix].Negotiate.HighestRoll<entryn) then
				if (dbg) then dbg.Log(prefix,"High roll received: "..entryn.." ("..sender..") - Previous high: "..self.DB[prefix].Negotiate.Roll); end
				self.DB[prefix].Negotiate.HighestRoll=entryn;
			end

		--[[ Stamp tag/marker ]]
		elseif (tag==DUCKNET_COMMAND.."M") then
			stampmarker=entry;
			if (stampmarker:match("^DMDBLIB.+")) then
				self.DB[prefix].Negotiate.SelfStamp=DM.Database.Util:GetStamp(prefix,stampmarker);
			else
				self.DB[prefix].Negotiate.SelfStamp=self.DB[prefix].CallBack.CheckStamp(stampmarker);	-- Get my addon's stamp
			end
		end

		--[[ Incoming entry sequence (numeric) ]]
		if (tag==DUCKNET_COMMAND.."E") then
			entry=entryn;
			self.Timers.LastEntry=now;
			if (entry~=self.DB[prefix].LastEntry+1) then												-- Sequence is broken
				self:Out(prefix,DUCKNET_COMMAND.."A"..DUCKNET_REQ_RETRANSMIT);		-- Request the data retransmitted
			else
				self.DB[prefix].LastEntry=self.DB[prefix].LastEntry+1;							-- Increase sequence
			end

		--[[ DuckNet handling ]]
		elseif (tag==DUCKNET_COMMAND.."A" and entry==DUCKNET_REQ_RETRANSMIT) then					-- An addon received inconsistency and requires a re-transmit
			if (self.DB[prefix].Transmitting>DUCKNET_TRANS_STOP) then			-- I am currently transmitting
				self.DB[prefix].Transmitting=DUCKNET_TRANS_START;				-- Start from the top
				self.DB[prefix].TransmitDelay=2;								-- 2 seconds delay for retransmission
				if (dbg) then dbg.Log(prefix,"Request for transmission restart from: "..sender); end
			end

		--[[ Someone is gonna send stuff ]]
		elseif (tag==DUCKNET_COMMAND.."A" and entry==DUCKNET_ACT_TRANSMIT) then
			self:StopNegotiation(prefix);
			self.Timers.LastEntry=now;
			self.DB[prefix].LastEntry=0;									-- Nothing received yet
			self.DB[prefix].Data={ };										-- Clear input buffer
			tp=self.DB[prefix].Data;					-- Resave as the able in new/changed
			if (dbg) then dbg.DN_ACT_TRANSMIT(prefix,sender,stampmarker,linestamp); end
			if (self.DB[prefix].Transmitting>=DUCKNET_TRANS_START) then		-- I am transmitting too
				if (DM.Database.Addon[prefix]) then
					DM.Database.Addon[prefix].PollingKey=nil;
					DM.DatabaseQueue[prefix].running=nil;
				end
				self.DB[prefix].Transmitting=DUCKNET_TRANS_STOP;			-- Stop transmission as several is trying to send
				self.DB[prefix].TransmitDelay=DM:Random(1,3);				-- Short delay
				if (dbg) then dbg.print(prefix,DM.Color.Red.."Some other are sending while I am sending (or I am about to send). Aborting my transmission.|r"); end;
				self.DB[prefix].CallBack.TX(self.DB[prefix].Transmitting,instance);	-- Notify addon about abort
			end

		--[[ Inbound transmission is finished ]]
		elseif (tag==DUCKNET_COMMAND.."A" and entry==DUCKNET_ACT_TRANSMITDONE) then
			if (self.DB[prefix].DBInfo["M"]:match("^DMDBLIB.+")) then
				if (self.DB[prefix].DBInfo["M"]:match("^DMDBLIBpipe")) then
					self.DB[prefix].LastEntry=-1;							-- not receiving anymore (ugly)
					DM.Database.Util.InData_pipe(prefix,self.DB[prefix].Data,sender);
				else
					local lTree,lKey=DM.Database.Util:ReceiveValue(prefix,self.DB[prefix].DBInfo["S"],self.DB[prefix].DBInfo["M"],self.DB[prefix].Data);
					if (lTree and prefix:match(".+_pipe$")) then				-- new data received through pipe
						DM.Database.Util:QueueForSend(nil,lTree,lKey,true);	-- run local-only update
					end
				end
			else
				if (not DM.WoWnet.Session:InDataIntercept(prefix,self.DB[prefix].Receiver,sender,self.DB[prefix].Data)) then	-- If not wownet
					self.DB[prefix].CallBack.RX(self.DB[prefix].Data,sender);	-- Give table-pointer to registered addon
				end
			end
			self.DB[prefix].LastEntry=-1;										-- Nothing received yet
			if (dbg) then dbg.DN_ACT_TRANSMITDONE(prefix,sender,self.DB[prefix].DBInfo["M"]); end

		--[[ An addon has newer data than someone else ]]
		elseif (tag==DUCKNET_COMMAND.."A" and entry==DUCKNET_ACT_MYSTAMP) then	-- Some addon informs about it's stamp
			if (dbg) then dbg.DN_Negotiation(prefix,sender,stampmarker,linestamp,lastroll); end
			if (linestamp>self.DB[prefix].Negotiate.HighestStamp) then			-- New stamp is highest of all
				self.DB[prefix].Negotiate.HighestStamp=linestamp;				-- Save it for test after negotiation timer
				self.DB[prefix].Negotiate.HighestRoll=lastroll;					-- Higher stamp, so set it's roll as highest
			end
			self.DB[prefix].Negotiate.HoldOff=DUCKNET_NEG_RUNNING;				-- Restart negotiation timer
			if (stampmarker:match("^DMDBLIB.+")) then
				self.DB[prefix].Negotiate.SelfStamp=DM.Database.Util:GetStamp(prefix,stampmarker);
			else
				self.DB[prefix].Negotiate.SelfStamp=self.DB[prefix].CallBack.CheckStamp(stampmarker);	-- Someone sends their stamp, so get my addon's stamp
			end
			if (self.DB[prefix].Negotiate.SelfStamp>self.DB[prefix].Negotiate.HighestStamp) then
				self.DB[prefix].Negotiate.Start=true;							-- Restart negotiating
				if (dbg) then dbg.Log(prefix,"ACT_MYSTAMP -> Lower stamp received. Restarting negotiation timer."); end
			end

		--[[ An addon wanna know whazzup ]]
		elseif (tag==DUCKNET_COMMAND.."A" and entry==DUCKNET_REQ_NEWDATA) then
			if (dbg) then dbg.DN_REQ_NEWDATA(prefix,sender,stampmarker,linestamp); end
			if (self.DB[prefix].Negotiate.Start or self.DB[prefix].Negotiate.HoldOff~=DUCKNET_NEG_STOPPED) then
				if (DM.Database.Addon[prefix]) then
					DM.Database.Addon[prefix].PollingKey=nil;
					DM.DatabaseQueue[prefix].running=nil;
				end
				if (dbg) then dbg.print(prefix,DM.Color.Red.."Someone is polling during negotiation. Stopping all current negotiation."); end
				self:StopNegotiation(prefix);
			else
				local thestamp;
				if (stampmarker:match("^DMDBLIB.+")) then
					thestamp=DM.Database.Util:GetStamp(prefix,stampmarker);
				else
					thestamp=self.DB[prefix].CallBack.CheckStamp(stampmarker);
				end
				self:StartNegotiation(prefix,stampmarker,thestamp,linestamp);
			end

			--[[ Commands to send to calling addon ]]
		elseif (tag:sub(1,1)==DUCKNET_COMMAND) then			-- Any command other than A E
			local tagtype=tag:sub(2,2);						-- Get the letter
			self.DB[prefix].Info[infoline]={};						-- Make (and clear) table
			self.DB[prefix].Info[infoline].Type=tagtype;			-- Save the command type
			if (tag==DUCKNET_COMMAND.."S") then
				self.DB[prefix].Info[infoline].Command=entryn;		-- Save the numeric command contents
			else
				self.DB[prefix].Info[infoline].Command=entry;		-- Save the command contents
			end
			self.DB[prefix].DBInfo[tagtype]=self.DB[prefix].Info[infoline].Command;
			infoline=infoline+1;
			if (dbg) then dbg.Tag(prefix,sender,tag,entry); end
		end


--[[	Incoming table data 	]]
		if (string.sub(tag,1,1)==DUCKNET_ENTRY) then			-- >> Entry
			if (string.sub(tag,2,2)=="n") then					-- The entry-name is numeric
				entry=entryn;									-- Convert it to a number
			end
			if (data) then
				if (not self:Numeric(type(data))) then
					if (string.find(data,DUCKNET_BOOLVALDATA.."true")==1) then data=true;		-- Set boolean 'true'
					elseif (string.find(data,DUCKNET_BOOLVALDATA.."false")==1) then data=false;	-- Set boolean 'false'
					else 																		-- Recreate string
						data=self:SwapText(data,DUCKNET_LINEBREAKhtml,DUCKNET_LINEBREAKnative);
						data=self:SwapText(data,DUCKNET_QUOTATION,"\"");
					end
				end
				tp[entry]=data;									-- There's data, so store it in the entry as a label
			else												-- There's no data, so we're building (or adding to) a table below
				if (not tp[entry]) then tp[entry]={}; end		-- Generate empty table
				tp=tp[entry];									-- Set new table pointer
			end;
		end
	end   -- Text-loop

	-- The line contained info, so tell the addon
	if (infoline>1 and self.DB[prefix].CallBack.Info) then
		self.DB[prefix].CallBack.Info(self.DB[prefix].Info);
	end
end

--MDMDBLIBpipe
--ADuckNet-Act-Transmit-Data
--E1
--n1"FIND_PIPERFun TimesLaughing Skull Clan"
--ADuckNet-Act-Transmit-Data-Done


function DM.Net:MakeBool(value)
	if (type(value)=="boolean") then return value; end
	if (not value) then return nil; end
	if (type(value)=="string") then value=tonumber(value); end
	if (type(value)=="number") then
		if (value==0) then return nil; end
		return true;
	end
	return nil;
end


-- Entry-builder
function DM.Net:MakeEntry(code,entry,data)
	if (not code) then return entry; end				-- Just return it
	if (not entry) then entry=tonumber(0); end			-- Just make sure it's a number

	local text="";
	local numeric=self:Numeric(type(entry));

	if (code==DUCKNET_ENTRY) then						-- It's an entry
		text=text..DUCKNET_ENTRY;						-- Set entry marker
		if (numeric) then text=text.."n";				-- It's a numeric entry
		else text=text.."s"; end						-- It's a string entry
	else text=text..DUCKNET_COMMAND..code; end			-- It's (probably) a command

	text=text..entry;									-- Add entry for supplied code

	if (code==DUCKNET_ENTRY and data) then				-- The entry carries data
		numeric=self:Numeric(type(data));				-- Check numeric
		local bool=self:Bool(type(data));				-- Check boolean
		text=text..DUCKNET_DATA;						-- Denote data
		if (not numeric) then							-- A string, so enclose
			text=text.."\"";
			if (bool) then
				if (data==true) then data=DUCKNET_BOOLVALDATA.."true"; else data=DUCKNET_BOOLVALDATA.."false"; end
			else
				data=self:SwapText(data,DUCKNET_LINEBREAKnative,DUCKNET_LINEBREAKhtml);	-- Transmittable linebreaks
				data=self:SwapText(data,"\"",DUCKNET_QUOTATION);
			end
		end
		text=text..data;																-- Add the entry data
		if (not numeric) then text=text.."\""; end										-- A string, so enclose
	end

	return text;
end


-- Handle further transmission
function DM.Net:SendNextLine(prefix)
	if (not self.DB[prefix].Output.Buffer[self.DB[prefix].Transmitting]) then		-- No more lines
		if (self.DB[prefix].PB) then self.DB[prefix].PB:SetValue(0); end
		local lines=self.DB[prefix].Transmitting;
		self.DB[prefix].Transmitting=DUCKNET_TRANS_STOP;								-- Not transmitting
		self.DB[prefix].CallBack.TX(lines);											-- Notify addon
		return;
	end

	if (self.DB[prefix].PB) then self.DB[prefix].PB:SetValue(self.DB[prefix].Transmitting); end
	self:Out(prefix,self.DB[prefix].Output.Buffer[self.DB[prefix].Transmitting]);		-- Send it
	self.DB[prefix].Transmitting=self.DB[prefix].Transmitting+1;												-- Go to next
end


-- Generic addon output
function DM.Net:Out(prefix,text)
	text=self:SwapText(text,DUCKNET_FROMCODE,DUCKNET_TOCODE);

	if (string.len(text)>254) then
		DM:Chat("Current line is too long ("..string.len(text)..")",1,0,0);
		return;
	end

	DM.Talk[prefix]=time();
-- Comment out these two to not get kicked at trans-debug
	SendAddonMessage(prefix,text,self.DB[prefix].CType,self.DB[prefix].Receiver);
	if (not self.DB[prefix].Receiver) then
		self.DB[prefix].LastOutput=text;
	end
end


--[[				   ]]
--[[ Support-functions ]]
--[[				   ]]


-- This function may seem a little overkill for a trained LUA programmer, but
-- better safe than sorry in case of any spec-changes
function DM.Net:SwapText(text,from,to)
	if (text) then
		while (string.find(text,from)) do text=string.gsub(text,from,to); end
	end
	return text;
end

-- Yeye... I know LUA natively does this (mostly), but this one restricts it a bit
function DM.Net:Numeric(typestring)
	if (typestring=="number" or typestring=="nil") then return true; end
	return nil;
end
function DM.Net:Bool(typestring)
	if (typestring=="boolean") then return true; end
	return nil;
end


--[[																  ]]
--[[ Please note: The following 2 functions are still at a beta stage ]]
--[[																  ]]

function DM.Net:AutoTableLinebreak(prefix,startnext)
	self:NewLine(prefix);													-- Advance

	if (table.maxn(self.DB[prefix].Output.Tables)<1) then
		if (startnext) then
			self.DB[prefix].Output.Buffer[self.DB[prefix].Output.Line]=DUCKNET_COMMAND.."E"..self.DB[prefix].Output.Entry;		-- !!! tp !!!
			self.DB[prefix].Output.Entry=self.DB[prefix].Output.Entry+1;
		end
	else
		for i,k in pairs(self.DB[prefix].Output.Tables) do							-- Iterate previous subtables
			if (not self:AddEntry(prefix,k,nil,true)) then return nil; end	-- Re-add them at next line
		end
	end
end


function DM.Net:CodeTable(prefix,t,level)
	local i,v;
	for i,v in pairs(t) do
		if (level==0) then self:NewLine(prefix); end							-- We're at the root and starting new entry, so make a new line
		if (type(v)=="table") then
			if (not self:AddTable(prefix,i)) then return nil; end				-- Add name of sub-table
			if (not self:CodeTable(prefix,v,level+1)) then return nil; end	-- Code sub-table
		else
			if (not self:AddEntry(prefix,i,v)) then return nil; end			-- Standard entry, so just add it with data
		end
	end

	table.remove(self.DB[prefix].Output.Tables);			-- Remove last table (default is last entry)
	self:AutoTableLinebreak(prefix);

	return true;
end


--[[  Checks if DuckNet is busy ]]
function DM.Net:NotCrusial(prefix)
	if (not self:Valid(prefix)) then return nil; end;									-- No such prefix
	if (self.DB[prefix].LastEntry>-1) then return nil; end;								-- I am currently receiving
	if (self.DB[prefix].Transmitting>=DUCKNET_TRANS_START) then return nil; end;		-- I am currently transmitting
	if (self.DB[prefix].Negotiate.Start) then return nil; end;							-- I am currently starting a negotiation
	return true;
end



--[[					]]
--[[	  User API	  ]]
--[[					]]


function DM.Net:ReturnNull()
	return 0;
end

function DM.Net:ConnectW(prefix,cbRX,receiver,persistent)
	if (not self:Connect(prefix,"WHISPER",nil,cbRX,nil,DM.WoWnet.Session.INFO,nil,nil,nil,nil)) then
		return nil;
	end
	self.DB[prefix].Receiver=receiver;
	if (not persistent) then self.DB[prefix].Alive=time()+120; end
	return true;
end

function DM.Net:Connect(prefix,...)
	local ctype,ctDATA,cbRX,cbTX,cbINFO,cbCS,cbNW,cbIS,pb=...;
	if (dbg) then dbg.Connect(prefix,...); end

	self:Disconnect(prefix);
	if (not (ctype=="PARTY" or ctype=="RAID" or ctype=="GUILD" or ctype=="BATTLEGROUND" or ctype=="WHISPER")) then
		return nil;
	end;

	-- Define the addon
	RegisterAddonMessagePrefix(prefix);
	self.DB[prefix]={ };
	self.DB[prefix].CType=ctype;
	self.DB[prefix].PB=pb;
	self.DB[prefix].LastOutput=nil;
	self.DB[prefix].LastEntry=-1;
	self.DB[prefix].Negotiate={ };
	self.DB[prefix].Negotiate.Start=false;
	self.DB[prefix].Negotiate.HoldOff=DUCKNET_NEG_STOPPED;
	self.DB[prefix].Negotiate.SelfStamp=0;
	self.DB[prefix].Negotiate.HighestStamp=0;
	self.DB[prefix].Negotiate.HighestRoll=0;
	self.DB[prefix].Negotiate.Roll=-1;
	self.DB[prefix].CallBack={ };
	self.DB[prefix].CallBack.RX=cbRX;
	self.DB[prefix].CallBack.TX=cbTX;
	self.DB[prefix].CallBack.Info=cbINFO;
	self.DB[prefix].CallBack.CheckStamp=cbCS;
	self.DB[prefix].CallBack.NegotiateWon=cbNW;
	self.DB[prefix].CallBack.InStamp=cbIS;

	-- Set up common data-space
	self.DB[prefix].Common={};
	self.DB[prefix].Common.Data=ctDATA;
	self.DB[prefix].Common.Boot=30;

	-- Prep some tables
	self.DB[prefix].Info={ };
	self.DB[prefix].Data={ };
	self.DB[prefix].Output = { };
	self.DB[prefix].Output.Buffer = { };
	self.DB[prefix].Output.Tables = { };

	-- Fake some callbacks if not provided
	if (not self.DB[prefix].CallBack.CheckStamp) then self.DB[prefix].CallBack.CheckStamp=self.ReturnNull; end
	if (not self.DB[prefix].CallBack.TX) then self.DB[prefix].CallBack.TX=self.ReturnNull; end
	if (not self.DB[prefix].CallBack.RX) then self.DB[prefix].CallBack.RX=self.ReturnNull; end

	-- Make usable
	self.DB[prefix].TransmitDelay=0;						-- No delay
	self.DB[prefix].Transmitting=DUCKNET_TRANS_STOP;		-- Not transmitting
	self.DB[prefix].Idle=true;								-- Ready to work now

	self:StopNegotiation(prefix);				-- NEW
	self:ClearOutput(prefix);					-- NEW

	return true;
end

--function DuckNet_UnRegister(prefix)
function DM.Net:Disconnect(prefix)
	self.DB[prefix]=nil;
end


function DM.Net:Status(prefix,ctype)
	local text="";
	if (not ctype) then ctype="GUILD"; end

	-- Valid
	if (self:Valid(prefix,ctype)) then text=text..DM.Color.Green;
	else text=text..DM.Color.Red; end
	text=text.."Net ";
	if (not self:Valid(prefix,ctype)) then return text; end
	-- Receiving
	if (self.DB[prefix].LastEntry>-1) then text=text..DM.Color.Red;
	else text=text..DM.Color.Green; end
	text=text.."Rx ";
	-- Transmitting
	if (self.DB[prefix].Transmitting>=1) then text=text..DM.Color.Red;
	else text=text..DM.Color.Green; end
	text=text.."Tx ";
	-- Negotiate start
	if (self.DB[prefix].Negotiate.Start) then text=text..DM.Color.Red;
	else text=text..DM.Color.Green; end
	text=text.."NegS ";
	-- Negotiate
	if (self.DB[prefix].Negotiate.HoldOff<DUCKNET_NEG_DONE and self.DB[prefix].Negotiate.HoldOff>DUCKNET_NEG_STOPPED) then text=text..DM.Color.Red;
	else text=text..DM.Color.Green; end
	text=text.."Neg ";
	-- Rebuilding
	if (self.DB[prefix].Rebuilding) then text=text..DM.Color.Yellow;
	else text=text..DM.Color.Green; end
	text=text.."ReB ";

	-- Buffer
	text=text.."|rBuf:";
	if (self.DB[prefix].Output.Line==DUCKNET_TRANS_START) then text=text..DM.Color.Green;	-- Empty
	else text=text..DM.Color.Yellow; end
	text=text..self.DB[prefix].Output.Line;

	return text;
end


-- Return true: Data will be sent shortly
-- Return false: Data can't be sent now
-- Return nil: Data is unsendable
----DM.Net:SendTable(DM.WoWnet.prefix,data,location);
function DM.Net:SendTable(prefix,ADD_Data,ADD_BlockName,ADD_Stamp)
	if (not ADD_Data) then return nil; end
	if (not ADD_BlockName) then ADD_BlockName=prefix; end
	if (not ADD_Stamp) then ADD_Stamp=0; end
	if (type(ADD_Stamp)~="number") then return nil; end
	if (not self:CanTransmit(prefix)) then return false; end					-- Can't transmit now
	self:ClearOutput(prefix);
	self:AddKey(prefix,"S",ADD_Stamp);
	self:AddKey(prefix,"M",ADD_BlockName);
	self:AddKey(prefix,"A",DUCKNET_ACT_TRANSMIT);
	self:NewLine(prefix);
	if (not self:CodeTable(prefix,ADD_Data,0)) then return nil; end			-- Could not build output with this data
	return self:DoTransmission(prefix);
end

function DM.Net:SendDatabase(prefix,ADD_Data,ADD_BlockName,ADD_Stamp)
	if (not ADD_Data) then return nil; end
	if (not ADD_BlockName) then ADD_BlockName=prefix; end
	if (not ADD_Stamp) then ADD_Stamp=0; end
	if (type(ADD_Stamp)~="number") then return nil; end
	if (not self:CanTransmit(prefix)) then return false; end					-- Can't transmit now
	self:ClearOutput(prefix);
	self:AddKey(prefix,"S",ADD_Stamp);
	self:AddKey(prefix,"M",ADD_BlockName);
	self:AddKey(prefix,"A",DUCKNET_ACT_TRANSMIT);
	self:NewLine(prefix);

	-- Add empty tables for each nested database
	local mergedb=DM:CopyTable(ADD_Data._str_);				-- Own data
	for k in dpairs_dbNL(ADD_Data) do	mergedb[k]={}; end		-- Insert nested databases as empty

	if (not self:CodeTable(prefix,mergedb,0)) then return nil; end			-- Could not build output with this data
	return self:DoTransmission(prefix);
end


--[[  Checks if DuckNet is ready to handle your addon  ]]
function DM.Net:Idle(prefix)
	if (not self:NotCrusial(prefix)) then return nil; end;
	if (self.DB[prefix].Negotiate.HoldOff>DUCKNET_NEG_STOPPED) then return nil; end;	-- Negotiation is currently running
	return true;
end


--[[  Checks if DuckNet is ready to transmit  ]]
function DM.Net:CanTransmit(prefix)
	if (not self:NotCrusial(prefix)) then return nil; end;
	if (self.DB[prefix].Negotiate.HoldOff<DUCKNET_NEG_DONE and self.DB[prefix].Negotiate.HoldOff>DUCKNET_NEG_STOPPED) then return nil; end;		-- Negotiation is off or done
	return true;
end


--[[  Send an entire table directly (Still being designed, so please don't use this yet)  ]]
function DM.Net:SendRawTable(prefix,t)
	if (not self:CanTransmit(prefix)) then							-- Invalid or busy
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ SendRawTable|r"); end
		return nil;
	end;
	-- Insert generic header
	self:AddKey(prefix,"M","Table");
	self:AddKey(prefix,"A",DUCKNET_ACT_TRANSMIT);
	self:NewLine(prefix);

	-- Build transmission queue
	local level=0;												-- Set root
	DM:ClearTable(self.DB[prefix].Output.Tables);
	if (not self:CodeTable(prefix,t,level)) then
		-- Error when coding the table. Handle it.
		return;
	end

	-- Footer
	self:DoTransmission(prefix);
end


--[[  Poll network for newer data  ]]
function DM.Net:Poll(prefix,stamp,marker,Tversion)
	if (not self:ClearOutput(prefix)) then
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ Poll|r"); end
		return nil;
	end
	if (dbg) then dbg.DN_Poll(prefix,stamp,marker,Tversion); end
	if (not stamp) then stamp=0; end
	self:AddKey(prefix,"S",stamp);
	if (marker) then self:AddKey(prefix,"M",marker); end
	if (Tversion) then self:AddKey(prefix,"T",Tversion); end
	self:AddKey(prefix,"A",DUCKNET_REQ_NEWDATA);
	self:StartNegotiation(prefix,marker,stamp);
	return self:SingleTransmit(prefix);
end


--[[  Clear output buffer  ]]
function DM.Net:ClearOutput(prefix)
	if (not self:CanTransmit(prefix)) then					-- Invalid or busy
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ ClearOutput|r"); end
		return nil;
	end;
--	self.DB[prefix].Output.Buffer={ };						-- Clear table
	if (not self.DB[prefix].Output.Buffer) then self.DB[prefix].Output.Buffer={};
	else wipe(self.DB[prefix].Output.Buffer); end
	if (not self.DB[prefix].Output.Tables) then self.DB[prefix].Output.Tables={};
	else wipe(self.DB[prefix].Output.Tables); end
	self.DB[prefix].Output.Line=1;							-- Start at the top
	self.DB[prefix].Output.Entry=1;							-- Start at the top
	return true;											-- It's cleared
end


--[[  Advance line in output buffer  ]]
function DM.Net:NewLine(prefix)
	if (not self:CanTransmit(prefix)) then					-- Invalid or busy
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ NewLine|r"); end
		return nil;
	end;
	if (self.DB[prefix].Output.Buffer[self.DB[prefix].Output.Line]) then		-- Current line exists
		self.DB[prefix].Output.Line=self.DB[prefix].Output.Line+1;		-- Go to next line
	end
	return true;												-- It's cleared
end


--[[  Add a new key/action/marker  ]]
function DM.Net:AddKey(prefix,name,info)
	if (not self:CanTransmit(prefix)) then					-- Invalid or busy
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ AddKey|r"); end
		return nil;
	end
	if (string.len(name)~=1) then return nil; end;
	local tp=self.DB[prefix].Output.Buffer;
	if (not tp[self.DB[prefix].Output.Line]) then
		tp[self.DB[prefix].Output.Line]="";
	end
	tp[self.DB[prefix].Output.Line]=tp[self.DB[prefix].Output.Line]..DUCKNET_COMMAND..name..info;
	return true;												-- It's cleared
end


--[[  Add a new table name  ]]
function DM.Net:AddTable(prefix,name)
	return self:AddEntry(prefix,name);
end


--[[  Add a new entry with data  ]]
function DM.Net:AddEntry(prefix,name,data,rebuild,nosplit)
	if (not self:CanTransmit(prefix)) then					-- Invalid or busy
		if (dbg) then dbg.print(prefix,DM.Color.Red.."DuckNet is not idle @ AddEntry|r"); end
		return nil;
	end
	if (string.len(name)<1) then return nil; end;				-- Name too short
	local tp=self.DB[prefix].Output.Buffer;
	if (not tp[self.DB[prefix].Output.Line]) then
		tp[self.DB[prefix].Output.Line]=DUCKNET_COMMAND.."E"..self.DB[prefix].Output.Entry;
		self.DB[prefix].Output.Entry=self.DB[prefix].Output.Entry+1;
	end

	-- Add it
	local lastline=tp[self.DB[prefix].Output.Line];
	local entry=self:MakeEntry(DUCKNET_ENTRY,name,data);
	tp[self.DB[prefix].Output.Line]=tp[self.DB[prefix].Output.Line]..entry;

	-- Check length. Break if need be
	local max=250-string.len(prefix);
	if (string.len(tp[self.DB[prefix].Output.Line])>max) then					-- Too long
		if (rebuild) then return nil; end											-- Currently rebuilding, data is untransmittable
		if (nosplit) then																	-- Can't split line
			self:AutoTableLinebreak(prefix,true);										-- Break and rebuild and start next line either way
			tp[self.DB[prefix].Output.Line]=tp[self.DB[prefix].Output.Line]..entry;
			if (string.len(tp[self.DB[prefix].Output.Line])>max) then return nil; end	-- Still too long - Can't rebuild
		else
			local overhead=string.len(DUCKNET_ENTRYBREAKCONTINUE);							-- Splitter overhead
			while (string.len(tp[self.DB[prefix].Output.Line])>max) do					-- Still too long
				if (not self:NewLine(prefix)) then return nil; end						-- Can't add line
				tp[self.DB[prefix].Output.Line]=tp[self.DB[prefix].Output.Line-1];	-- Copy before crop-clip
				tp[self.DB[prefix].Output.Line-1]=string.sub(tp[self.DB[prefix].Output.Line],1,max-overhead)..DUCKNET_ENTRYBREAKCONTINUE;
				local UTF=0;
				while (string.len(tp[self.DB[prefix].Output.Line-1])>max) do		-- Bleedin' charcoding...
					tp[self.DB[prefix].Output.Line-1]=string.sub(tp[self.DB[prefix].Output.Line],1,(max-overhead)-UTF)..DUCKNET_ENTRYBREAKCONTINUE;
					UTF=UTF+1;
				end
				tp[self.DB[prefix].Output.Line]=string.sub(tp[self.DB[prefix].Output.Line],((max-overhead)-UTF)+1);
			end
		end
	end
	if (rebuild) then return true; end									-- Rebuild complete
	-- At this point, any requested data has been added successfully

	-- Table book-keeping
	if (not data) then table.insert(self.DB[prefix].Output.Tables,name); end	-- Add this table

	return true;												-- It's cleared
end


--[[  Finalise and start transmission  ]]
function DM.Net:DoTransmission(prefix,freeform)
	if (not freeform) then
		if (not self:NewLine(prefix)) then
			if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not add line @ DoTransmission|r"); end
			return nil;
		end
		if (not self:AddKey(prefix,"A",DUCKNET_ACT_TRANSMITDONE)) then return nil; end
	end

	-- Validate buffer
	local tp=self.DB[prefix].Output.Buffer;
	local overhead=string.len(prefix)+1;
	local seq=1;
	local linebreaks=0;
	local longlines=0;
	while (tp[seq]) do
		if (string.find(tp[seq],DUCKNET_LINEBREAKnative)) then
			linebreaks=linebreaks+1;
			DM:Chat("ERROR: Linebreak in line "..seq,1,0,0);
			DM:Chat(tp[seq],1,0,0);
		end
		local total=string.len(tp[seq])+overhead;
		if (total>253) then longlines=longlines+1; DM:Chat("ERROR: Line "..seq.." length: "..total,1,0,0); end
		seq=seq+1;
	end
--print("Output buffer is "..(seq-1).." lines");
	if (seq==2) then return false; end
	if (linebreaks>0 or longlines>0) then
		self:ClearOutput(prefix);
		DM:Chat("ERROR: Transmission aborted",1,0,0);
		return true;
	end

	if (self.DB[prefix].PB) then
		self.DB[prefix].PB:SetMinMaxValues(0,self.DB[prefix].Output.Line);
		self.DB[prefix].PB:SetValue(0);
	end

	self.DB[prefix].Transmitting=DUCKNET_TRANS_START;		-- Start transmission
	self.DB[prefix].LastOutput=nil;							-- It's there now
	return true;
end


--[[  Send the first line only  ]]
function DM.Net:SingleTransmit(prefix)
	if (not self.DB[prefix].Output.Buffer[1]) then return nil; end;			-- The line is empty
	self:Out(prefix,self.DB[prefix].Output.Buffer[1]);	-- Send it
	return true;
end

function DM.Net:GoodDatatype(datatype)
	if (datatype=="number") then return true; end
	if (datatype=="string") then return true; end
	if (datatype=="boolean") then return true; end
	return nil;
end

function DM.Net:SendBlindWhisper(prefix,receiver,name,thedata,marker)
	if (self:Valid(prefix) and not DM.Net:CanTransmit(prefix)) then return nil; end	-- valid and it's busy
	local text="";
	if (marker) then text=text..DUCKNET_COMMAND.."M"..marker; end
	text=text..DUCKNET_COMMAND.."A"..DUCKNET_ACT_TRANSMIT;
	text=text..DUCKNET_COMMAND.."E1";
	text=text..self:MakeEntry(DUCKNET_ENTRY,name,thedata);
	text=text..DUCKNET_COMMAND.."A"..DUCKNET_ACT_TRANSMITDONE;
	text=self:SwapText(text,DUCKNET_FROMCODE,DUCKNET_TOCODE);
	if (text:len()>254) then
		DM:Chat("Current line is too long @ SendBlindWhisper ("..text:len()..")",1,0,0);
		return nil;
	end
	if (not DM.DefaultChatHook_AddMessage_Original) then
        DM.DefaultChatHook_AddMessage_Original=DEFAULT_CHAT_FRAME.AddMessage;
        DEFAULT_CHAT_FRAME.AddMessage=DM.DefaultChatHook_AddMessage;
    end
	DM.SentBlindWhisperListOfReceivers[receiver]=time();
	SendAddonMessage(prefix,text,"WHISPER",receiver);
	return true;
end

--[[
function DM.Net:SendSimple(prefix,name,thedata,marker)
	if (not name) then return nil; end
	if (not thedata) then return nil; end
	if (not self:GoodDatatype(type(thedata))) then return nil; end
	if (not self:ClearOutput(prefix)) then return nil; end
	if (marker) then
		if (not self:AddKey(prefix,"M",marker)) then return nil; end
	end
	if (not self:AddKey(prefix,"A",DUCKNET_ACT_TRANSMIT)) then return nil; end
	if (not self:NewLine(prefix)) then return nil; end
	if (not self:AddEntry(prefix,name,thedata)) then return nil; end
	return self:DoTransmission(prefix);
--	local ret=self:DoTransmission(prefix);
--	if (ret==false) then return self:SendSimple(prefix,name,thedata,marker); end
--	return ret;
end
]]

-- This is a bit simplified, as it does not have support for rendering
-- more than one page at the time. This is however regarded a minor problem,
-- as rendering a page is two-pass with a single update-frame inbetween.
-- The rest of the code do however support multiple simultaneous pages.
-- Check the return value for success. "nil" is fail. "true" is victory.
-- If "nil", retry next update-frame.
function DM:Render(frame,sectiondata)
	if (DM.RU.Canvas) then return nil; end	-- Could not render at this time
	if (type(sectiondata)=="table") then			-- SECTIONDATA
	elseif (type(sectiondata)=="string") then		-- SECTIONDATA.Data
		sectiondata={
			Data=sectiondata,
			Keys="",
			Description="Anonymous data",
			Meta={
				Server="",
				Location="",
			},
		};
	else return; end

	DM.RU.Canvas=frame;
	DM.RU.Canvas.SECTIONDATA=sectiondata;
	if (not frame.WnRenderData) then
		frame.WnRenderData={};
		frame.WnRenderData.Widget={};
		frame.WnRenderData.Focus={};
		frame.WnRenderData.WnContext={};
		frame.GoTo=DM.RU.SetCanvasFocus;
	end

	DM.RenderWnCounter=0;
	DM.RenderWnRemaining=sectiondata.Data;
	if (DM.RU.Canvas:GetParent().SetVerticalScroll) then DM.RU.Canvas:GetParent():SetVerticalScroll(0); end
	if (DM:Render_2(DM.RU.Canvas,DM.RenderWnRemaining)) then return true; end
	-- Fail, so do fail-stuff
	DM.RU.Canvas=nil;
	DM.RenderWnCounter=1000;
	return nil;
end

function DM:Render_2(canvas,remaining)
	DM.RenderWnCounter=DM.RenderWnCounter+1;
	canvas.WnRenderData.LockH=0;			-- Far left
--DM:Chat("Rendering...",1,0,1);
	-- Do blocks
	local entryname,param,data;
--	entryname,param,data,remaining=DM.RU:PullSection(remaining);
	entryname,param,remaining=DM.RU:PullSection(remaining);
	if (entryname~="<wn>") then return; end			-- Wrong data-type waiting
	DM.RU:InitPage(canvas);	-- This will work, as the FontString widget will be set properly before a new redraw
	while (remaining~="") do
		entryname,param,data,remaining=DM.RU:PullSection(remaining);
		if (entryname=="<meta>") then
--DM:Chat("Setting up meta...",1,0,1);
			local remM=data;						-- Do the meta data
			while (remM~="") do
				entryname,param,data,remM=DM.RU:PullSection(remM);
				if (entryname=="<elements>") then
					local remE=data;				-- Do the elements data
					while (remE~="") do
						entryname,param,data,remE=DM.RU:PullSection(remE);
						-- Overlay parameters
						canvas.WnRenderData.Meta["<elements>"][entryname]=DM.RU:Mix(param,canvas.WnRenderData.Meta["<elements>"][entryname]);
						DM.RU:SetInterface(canvas,entryname,canvas.WnRenderData.Meta["<elements>"][entryname].Class)
					end
				elseif (entryname=="<font>") then
					local name="Default";
					if (param.name) then name=param.name; end
					param.name=nil;
					canvas.WnRenderData.Meta.font[name]=DM:CopyTable(param,canvas.WnRenderData.Meta.font[name]);
				else
					canvas.WnRenderData.Meta[entryname]=DM:CopyTable(param,canvas.WnRenderData.Meta[entryname]);
				end
			end
		elseif (entryname=="<body>") then
			if (not canvas.MeasureFontString) then
				canvas.MeasureFontString=DM.RU:GetWidget(canvas,"<stext>",canvas.WnRenderData.Meta["<elements>"]["<stext>"],nil);
				if (not canvas.MeasureFontString) then
					DM:Chat("Could not create a DuckMod measurement frame",1);
					return;
				end
				canvas.MeasureFontString.SPECIALNOREUSE=true;
			end
--DM:Chat("Rendering body...",1,0,1);
			-- Set the page standards from meta
			if (canvas.WnRenderData.Meta["<page>"]) then
				if (canvas.WnRenderData.Meta["<page>"].background) then
					local bgwidget=DM.RU:GetWidget(canvas,"<img>",canvas.WnRenderData.Meta["<elements>"]["<img>"],nil);
					bgwidget:SetDrawLayer("BACKGROUND");
					bgwidget:ClearAllPoints();
					bgwidget:SetPoint("TOPLEFT",canvas,"TOPLEFT");
					bgwidget:SetPoint("BOTTOMRIGHT",canvas,"BOTTOMRIGHT");
					local r,g,b,a=DM.RU:SplitColor(canvas.WnRenderData.Meta["<page>"].background);	-- Try to make RGBA
					if (r) then bgwidget:SetTexture(r,g,b,a);							-- Single color
					else bgwidget:SetTexture(canvas.WnRenderData.Meta["<page>"].background); end		-- Image-file
					bgwidget:Show();
				end
			end
			-- Do the contents from data
			local remB=data;						-- Do the body data
			while (remB~="") do
				entryname,param,data,remB=DM.RU:PullSection(remB);
				if (not entryname) then
					DM:Chat("Unknown tag: ?",1);
				elseif (entryname=="<set>") then
					DM.RU:SetContext(canvas,param);	-- Non-visual elements
				elseif (entryname=="<focus>") then
					if (param.id) then canvas.WnRenderData.Focus[param.id]=canvas.WnRenderData.Bottom; end
				elseif (not canvas.WnRenderData.Meta["<elements>"][entryname]) then
					DM:Chat("Unknown tag: "..entryname,1);
				else
					-- Visual elements
					if (not canvas.WnRenderData.Meta["<elements>"][entryname]) then DM:Chat("Missing element <"..entryname..">",1); return; end
					-- * Create a widget, or retrieve an old from a stack.
					local widget=DM.RU:GetWidget(canvas,entryname,canvas.WnRenderData.Meta["<elements>"][entryname],param);
					-- * Make parameters
					local mix=DM.RU:Mix(param,canvas.WnRenderData.Meta["<elements>"][entryname]);
					-- * Set provided data
					local r,g,b,a=DM.RU:SplitColor(data);	-- Try to make an RGBA
					if (not r) then
						if (DM.RU:Flag(canvas,canvas.WnRenderData.Meta["<elements>"][entryname],"FontInstance")) then
							DM.RU:SetFont(canvas,widget,mix);
						end
						widget[canvas.WnRenderData.Meta["<elements>"][entryname].IF.SetData](widget,data);	-- Simulate colon
					else
						widget[canvas.WnRenderData.Meta["<elements>"][entryname].IF.SetData](widget,r,g,b,a);	-- Simulate colon
					end
					-- * Set it for placeholder if needed
					local swap=DM.RU:Flag(canvas,canvas.WnRenderData.Meta["<elements>"][entryname],"SwapToLink");
					if (swap) then
						DM.RU:InitWidget(canvas.MeasureFontString);
						DM.RU:CopyTextAttributes(widget,canvas.MeasureFontString);
						canvas.MeasureFontString:SetText(data);
					end
					-- * Anchor and size
					DM.RU:SetPosition(canvas,widget,mix,DM.RU:Flag(canvas,canvas.WnRenderData.Meta["<elements>"][entryname],"SwapToLink"));
					-- * Set it for the actual widget if needed
					if (swap) then
						widget[canvas.WnRenderData.Meta["<elements>"][entryname].IF.SetData](widget,data);	-- Simulate colon
					end
--					if (widget.SetBackdrop) then
--						local backdrop={bgFile = "Interface\\DialogFrame\\UI-DialogBox-Gold-Background",edgeFile=nil,tile=false,tileSize=1,edgeSize=0,insets={left=0,right=0,top=0,bottom=0}}
--						widget:SetBackdrop(backdrop);
--					end
					-- * Show it
					widget:Show();
					canvas.WnRenderData.LastWidget=widget;
				end
			end
			canvas:SetHeight(canvas.WnRenderData.Bottom+1);
		else
			-- Unknown block
		end
	end
	DM.RenderWnCounter=DM.RenderWnCounter+1;

	-- Finalise the page
	if (DM.RenderWnCounter==4) then
		DM.RU.Canvas=nil;						-- Done with this canvas
	end
--	if (DM.RenderWnCounter==2) then
--		DM.RenderWnCounter=4;
--		DM.RU.Canvas=nil;						-- Done with this canvas
--	end
	return true;
end


-- Render Utilities
DM.RU={
	Class={
		-- "=" denotes the named class
		-- "x" denotes included entities
		HaveContainer = 0x00000001, -- x x
		UIObject	  = 0x00000002, -- x x x x x x x
		FontInstance  = 0x00000004, -- x   x	 x x
		Region		= 0x00000008, -- x x x x x x x
		LayeredRegion = 0x00000010, -- x x
		FontString	= 0x00000020, -- =
		Texture	   = 0x00000040, --   =
		Frame		 = 0x00000080, --	 x x = x x
		EditBox	   = 0x00000100, --	 =
		Button		= 0x00000200, --	   =
		SimpleHTML	= 0x00000400, --		   =
		ScrMsgFrame   = 0x00000800, --			 =
		SwapToLink	= 0x00001000, --
		DeadEditBox   = 0x00002000,

		WNCheckBox	= 0x40000000,
		WNListBox	 = 0x80000000,

		FontString_   = 0x0000003F, -- Full FontString
		Texture_	  = 0x0000005B, -- Full Texture
		Frame_		= 0x0000008A, -- Full Frame
		EditBox_	  = 0x0000018E, -- Full EditBox
		Button_	   = 0x0000028A, -- Full Button
		TextLink_	 = 0x0000083F, -- FontString + SwapToLink
		SimpleHTML_   = 0x0000148E, -- Full SimpleHTML + swap-flag
		ScrMsgFrame_  = 0x0000088E, -- Full ScrollingMessageFrame
		DeadEditBox_  = 0x0000318E, -- Full editBox with no interaction
	},
};


-- Deviations to the default set-up of parameters are done by table overlay.
DM.RU.DefaultMeta={
	-- Create the building-blocks
	["<page>"]={
		background="#FFFFFFFF",								-- #AARRGGBB or filename
--		package="",											-- A registered package
	},
	font={			-- EXCEPTION: Not bracketed, as it is not a tag-to-entry table
		Default={
			inheritSys="GameFontNormal",	-- Inherit a system font by name
			size="14",
			color="#FF000000",				-- Black opaque
			flags="",
		},
		Medium={
			size="22",
		},
		Big={
			size="28",
		},
		Edit={
			color="#FF505050",
		},
	},
	-- Create the elements
	["<elements>"]={
		["<none>"]={							-- dummy
			Class=0,
		},
		["<text>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			Class=DM.RU.Class.FontString_,
		},
		["<stext>"]={
			font="Default",
			pad=3,
			widget="FontString",				-- the widget
			Class=DM.RU.Class.FontString_,
		},
		["<ltext>"]={
			font="Default",
			pad=3,
			widget="EditBox",					-- the widget
			Class=DM.RU.Class.DeadEditBox_,
		},
		["<h1>"]={
			widget="FontString",				-- the widget
			font="Big",
			pad=3,
			Class=DM.RU.Class.FontString_,
		},
		["<h2>"]={
			widget="FontString",				-- the widget
			font="Medium",
			pad=3,
			Class=DM.RU.Class.FontString_,
		},
		["<img>"]={
			widget="Texture",					-- the widget
			pad=3,
			Class=DM.RU.Class.Texture_,
		},
		["<edit>"]={
			widget="EditBox",					-- the widget
			inheritSys="DuckMod_EditText_01",	-- Inherit a frame
			width=200,
			height=25,
			font="Edit",
			pad=3,
			Class=DM.RU.Class.EditBox_,
		},
		["<button>"]={
			widget="Button",					-- the widget
			inheritSys="UIPanelButtonTemplate2",
			width=125,
			height=22,
			pad=3,
			Class=DM.RU.Class.Button_,
		},
	},
};

-- Each canvas (frame) will have the following additional info:
-- canvas
--	WnRenderData
--	   Widget	 - Storage for widget objects
--	   Meta	   - Meta for this page. That is: compined default and provided.
--	   LastWidget - Last rendered widget
--	   Bottom	 - Current end-of-page
--	   WnContext  - Current page context (<set>)
--	   LockH	  - The current LockH while rendering

-- Since each canvas keeps its own registry, multiple rendered pages is supported.
function DM.RU:InitPage(canvas)
	canvas.WnRenderData.Meta=DM:CopyTable(DM.RU.DefaultMeta);	-- Set default meta
	canvas.WnRenderData.LastWidget=nil;				-- nil is the container frame, and true only for the first widget
	canvas.WnRenderData.Bottom=0;					-- Bottom of reendered page
	canvas.WnRenderData.LockH=0;					-- Last locked horizontal
	-- START: v2.9903
	canvas:SetScript("OnMouseDown",DM.RU.CanvasClicked);
--	if (not canvas.WnLinkMenu) then canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",canvas,"UIDropDownMenuTemplate"); end
	if (not canvas.WnLinkMenu) then
--		canvas.WnLinkMenuAnchor=CreateFrame("Frame","WnLinkerMenuDisplayPositioning",UIParent);
--		canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",canvas.WnLinkMenuAnchor,"UIDropDownMenuTemplate");
		canvas.WnLinkMenu=CreateFrame("Frame","WnLinkerMenuDisplay",UIParent,"UIDropDownMenuTemplate");
	end
	-- END
	wipe(canvas.WnRenderData.WnContext);			-- Current run-length context
	wipe(canvas.WnRenderData.Focus);				-- Clear foci
	-- Clear all used elements
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
		for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
			wwTable.Used=nil;						-- New page, so all unused
			DM.RU:InitWidget(wwTable.widget);
		end
	end

--DM:Chat("Setting default interfaces...",1,0,1);
	-- Set interface for all defaults
	for nElement,dElement in pairs(canvas.WnRenderData.Meta["<elements>"]) do
		self:SetInterface(canvas,nElement,dElement.Class)
	end
end

-- v2.9903
function DM.RU.CanvasClicked(frame,button,down)
	if (button~="LeftButton") then return; end
	for element,eTable in pairs(frame.WnRenderData.Widget) do
		for name,nTable in pairs(eTable) do
			if (nTable.widget and nTable.widget.IsMouseOver and nTable.widget.GetText) then
				if (nTable.widget:IsMouseOver()) then		-- We're hovering this widget
					-- Walk through all links if any
					local text=nTable.widget:GetText();
					local links={};
					local index=1;
					for link,atom in text:gmatch("(|c.-|H(.-:.-)|h.-|h|r)") do
						links[index]={ text=link, value=atom, func=DM.RU.LinkMenuClicked, notCheckable=1, }
						index=index+1;
					end

					if (next(links)) then
						frame.WnLinkMenu.WnData=nil;		-- Disconnect
						frame.WnLinkMenu.WnData=links;		-- Set new
						frame.WnLinkMenu.WnWidget=nTable.widget;
						frame.WnLinkMenu.WnButton=button;
						EasyMenu(links,frame.WnLinkMenu,"cursor",0,0,"MENU",3);
					end
					return;
				end
			end
		end
	end
end

-- v2.9903
function DM.RU.LinkMenuInitialise(self)
--	DM.RU:AddHeader("Links");
	for link,atom in pairs(self.WnData) do
		DM.RU:AddButton(link,DM.RU.LinkMenuClicked,atom,"");
	end
end

function DM.RU.LinkMenuClicked(self)
	local menu=_G["WnLinkerMenuDisplay"];
	DM.RU.Widget_OnHyperlinkClicked(menu.WnWidget,self.value,self.text,menu.WnButton);
end

--DM.Color.White="|cFFFFFFFF",
--DM.Color.Green="|cFF00FF00",
--DM.Color.Red="|cFFFF0000",
--DM.Color.lBlue="|cFF6060FF",
--DM.Color.hBlue="|cFEA0A0FF",
--DM.Color.Yellow="|cFFFFFF00",

-- v2.9903
function DM.RU:AddHeader(text,icon)
	local info=UIDropDownMenu_CreateInfo();
	info.text=DM.Color.lBlue..text;
	info.icon=icon;
	UIDropDownMenu_AddButton(info,1);
end
-- v2.9903
function DM.RU:AddButton(text,func,value,icon)
	local info=UIDropDownMenu_CreateInfo();
	info.text=DM.Color.White..text;
	info.func=func;
	info.value=value;
	info.icon=icon;
	info.notCheckable=1;
	UIDropDownMenu_AddButton(info,1);
end

function DM.RU:InitWidget(widget)
	widget:Hide();					-- Hide all
--	if (widget.SetText) then widget:SetText(""); end
	widget:ClearAllPoints();
	widget:SetHeight(0);
	widget:SetWidth(0);
end

DM.RU.ScrMsgFrame={};
function DM.RU.ScrMsgFrame.SetText(self,text,r,g,b)
	self:Clear();
	self:AddMessage(text,r,g,b);
end

DM.RU.SimpleHTML={};
function DM.RU.SimpleHTML._SetText(self,text)
	self.DMRUTextContents=text;
	self:SetText(text);
end
function DM.RU.SimpleHTML._GetText(self)
	return self.DMRUTextContents;
end

function DM.RU.SetCanvasFocus(self,id)
	if (not self.WnRenderData) then return; end
	if (not self.WnRenderData.Focus) then return; end
	if (not self.WnRenderData.Focus[id]) then return; end
	if (self:GetParent().SetVerticalScroll) then
		self:GetParent():SetVerticalScroll(self.WnRenderData.Focus[id]);
	end
end

--	DM:Chat("Clicked: "..linkdata);	-- essential center - "H" part
--	DM:Chat("Clicked: "..link);		-- Complete link text
--	DM:Chat("Clicked: "..button);	-- "LeftButton"
function DM.RU.Widget_OnHyperlinkClicked(widget,linkdata,link,button)
	DM.RU.SendEvent(widget,"DMEVENT_HYPERLINK_CLICKED",linkdata,link,button);
end
function DM.RU.Widget_OnHyperlinkEnter(widget,linkdata,link)
	DM.RU.SendEvent(widget,"DMEVENT_HYPERLINK_ENTER",linkdata,link);
end
function DM.RU.Widget_OnHyperlinkLeave(widget,linkdata,link)
	DM.RU.SendEvent(widget,"DMEVENT_HYPERLINK_LEAVE",linkdata,link);
end

function DM.RU.SendEvent(widget,event,arg1,arg2,arg3,arg4)
	local parent=widget:GetParent():GetParent():GetParent();	-- frame,scrollframe,appwindow
	local handler=parent:GetScript("OnEvent");
	if (not handler) then
		parent=widget:GetParent():GetParent();
		handler=parent:GetScript("OnEvent");
		if (not handler) then
			parent=widget:GetParent();
			handler=parent:GetScript("OnEvent");
			if (not handler) then DEFAULT_CHAT_FRAME:AddMessage("No handler"); return; end
		end
	end
	handler(parent,event,widget:GetParent(),arg1,arg2,arg3,arg4);
end

function DM.RU.PageButtonClicked(button)
	if (not button) then
		DM:Chat("No button provided");
	else
		local canvas=button.WnCanvas;
--DM:Chat("Button clicked: "..button:GetName());
		DM.RU.SendEvent(button,"DMEVENT_BUTTON_CLICKED",button.WnContext,button.WnParam);
		if (not button.WnParam) then						-- No params in XML
			DM:Chat("No action associated with this button.");
		elseif (button.WnParam.send) then						-- param "send" supplied
			if (button.WnContext[button.WnParam.send]) then		-- It's send-data exists in the context
				local val=button.WnContext[button.WnParam.send];
				local transData={};
				local stored=nil;
				for wEntry,wTable in pairs(canvas.WnRenderData.Widget) do	 -- All widget groups
					local GetMyData=canvas.WnRenderData.Meta["<elements>"][button.WnElement].IF.GetData;
					for wwEntry,wwTable in pairs(wTable) do		-- All widgets within group
						if (canvas.WnRenderData.Meta["<elements>"][wwTable.widget.WnElement].IF.GetData) then
							if (wwTable.Used and wwTable.widget.WnParam.id and wwTable.widget.WnContext[button.WnParam.send]==val) then
								transData[wwTable.widget.WnParam.id]=wwTable.widget[GetMyData](wwTable.widget);
								stored=true;
							end
						end
					end
				end
				-- We've got data, so send it
				if (stored) then
--					DM.Net:ConnectW(prefix,cbRX,receiver);
					local canvas=button:GetParent();
					if (not canvas.SECTIONDATA) then return; end			-- No meta-data
					if (not canvas.SECTIONDATA.Meta) then return; end			-- No meta-data
					if (not canvas.SECTIONDATA.Meta.Server) then return; end		-- No server to send to
					-- Make a queue for polling and sending depending on result
					DM.WoWnet.Session:SendPublic("ping:"..canvas.SECTIONDATA.Meta.Server);	-- Do ping
					if (not DM.WoWnet.Session.ResolveWait) then				-- Nothing else waiting
						DM.WoWnet.Session.ResolveWait={
							Table=transData,					-- Actual data to send
							Server=canvas.SECTIONDATA.Meta.Server,	-- Server to send to
							PipeToon=nil;						-- Who's the toon
						};
					end
				end
			end
		else
--			DM:Chat("No known action associated with this button.");
		end
	end
end


-- The following code enables context-code, like so:
-- <set form="guild-application"/>  -- Starting form for guild application (multiple forms allowed)
-- <edit req="1"/>				  -- Assigned to form "guild-application" (requires at least one character data)
-- <button send="form"/>			-- Assigned to form "guild-application"
-- <set form="guild-log-in"/>	   -- Starting form for log-in, guild application form is done
-- <edit ID="LIname"/>
-- <edit ID="LIpass"/>
-- <set form=""/>					-- No form active
-- Following this, pressing the button will collect data from all active elements
-- associated with the same context. In this case "form".
function DM.RU:SetContext(canvas,param)
	if (not param) then param={}; end
	canvas.WnRenderData.WnContext=DM.RU:Mix(param,canvas.WnRenderData.WnContext);
	for entry,eVal in pairs(canvas.WnRenderData.WnContext) do
		if (eVal=="") then canvas.WnRenderData.WnContext[entry]=nil; end		-- Remove emptied parameters
	end
end

function DM.RU:Mix(paramA,paramB)
	if (not paramA) then paramA={}; end
	local param=DM:CopyTable(paramB);		-- Make copy
	return DM:CopyTable(paramA,param);		-- Merge and return
end

function DM.RU:CopyTextAttributes(wF,wT)
	wT:SetFont(wF:GetFont());
	wT:SetSpacing(wF:GetSpacing());
end

-- Handle "glue" and "flow"
function DM.RU:SetPosition(canvas,widget,param,swap)
	if (not param) then param={}; end
	local glue,flow,align,pad=param.glue,param.flow,param.align,param.pad;		-- Make copies
	local bottomtype="stack";

	if (swap) then
--		DM.RU:CopyTextAttributes(widget,canvas.MeasureFontString)
		swap=widget;
		widget=canvas.MeasureFontString;
	end

	widget:ClearAllPoints();
	local cWidth=canvas:GetWidth();
	local hSet,wSet;
	if (param.width) then
		local value=tostring(param.width);
		if (value:sub(-1)=="%") then
			value=tonumber(value:sub(1,-2));			-- Get the percentage
			value=(cWidth/100)*value;	-- Convert to points
		end
--DM:Chat("Width: "..tonumber(value));
		widget:SetWidth(tonumber(value));				-- Set provided value directly
		wSet=true;
	end
	if (param.height) then
		widget:SetHeight(tonumber(param.height));		-- Set provided value directly
		hSet=true;
	end

	-- Check provided information
	if (not glue and not flow and not align) then align="LEFT"; end
	if (not align) then align=""; end
	if (not glue) then glue=""; end
	if (not flow) then flow=""; end
	if (not pad) then pad=0; end
	align=align:upper();
	glue=glue:upper();
	flow=flow:upper();

--DM:Chat("g:"..glue.." f:"..flow.." a:"..align);

	-- Set text adjustment. "align" has presedence as it does that for positioning as well.
	local adjust;
	adjust="LEFT";		-- Since we are left-to-right readers
	if (flow=="LEFT") then adjust="RIGHT"; end
	if (flow=="DOWN") then adjust="CENTER"; end
	if (align=="LEFT" or align=="RIGHT" or align=="CENTER") then adjust=align; end

	local lockside,bottomadd="LEFT",pad;
	-- Align with canvas in stead of widget
	if (align=="LEFT" or (not canvas.WnRenderData.LastWidget and glue=="LEFT")) then
		widget:SetPoint("TOPLEFT",canvas,"TOPLEFT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		canvas.WnRenderData.LockH=0;
		adjust="LEFT";
	elseif (align=="RIGHT" or (not canvas.WnRenderData.LastWidget and glue=="RIGHT")) then
		widget:SetPoint("TOPRIGHT",canvas,"TOPRIGHT",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="RIGHT";
		canvas.WnRenderData.LockH=cWidth-0;
		adjust="RIGHT";
	elseif (align=="CENTER" or (not canvas.WnRenderData.LastWidget and glue=="TOP")) then
		widget:SetPoint("TOP",canvas,"TOP",0,0-(canvas.WnRenderData.Bottom+bottomadd));
		lockside="CENTER";
		canvas.WnRenderData.LockH=cWidth/2;
		adjust="CENTER";
	elseif (glue=="LEFT") then
		if (flow=="DOWN") then widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"TOPRIGHT",pad,0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMLEFT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",pad,0); bottomtype="BOTTOMUP";
		else widget:SetPoint("LEFT",canvas.WnRenderData.LastWidget,"RIGHT",pad,0); bottomtype="CENTERSIDE"; end
		adjust="LEFT";
		bottomadd=0;
	elseif (glue=="RIGHT") then
		if (flow=="DOWN") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"TOPLEFT",-pad,0); bottomtype="TOPDOWN";
		elseif (flow=="UP") then widget:SetPoint("BOTTOMRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",-pad,0); bottomtype="BOTTOMUP";
		else widget:SetPoint("RIGHT",canvas.WnRenderData.LastWidget,"LEFT",-pad,0); bottomtype="CENTERSIDE"; end
		lockside="RIGHT";
		adjust="RIGHT";
		bottomadd=0;
	elseif (glue=="TOP") then
		if (flow=="DOWN") then widget:SetPoint("TOP",canvas.WnRenderData.LastWidget,"BOTTOM",0,0-pad); lockside="CENTER"; adjust="CENTER";
		elseif (flow=="LEFT") then widget:SetPoint("TOPRIGHT",canvas.WnRenderData.LastWidget,"BOTTOMRIGHT",0,0-pad); lockside="RIGHT"; adjust="RIGHT";
		else widget:SetPoint("TOPLEFT",canvas.WnRenderData.LastWidget,"BOTTOMLEFT",0,0-pad); lockside="LEFT"; adjust="LEFT"; end
	end
	if (widget.SetJustifyV) then widget:SetJustifyV("TOP"); end
	if (widget.SetJustifyH) then widget:SetJustifyH(adjust); end

	-- Expand widget if text is bigger than said area
	if (not wSet and widget.GetStringWidth) then
		local sWidth=widget:GetStringWidth();

		if (lockside=="LEFT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+pad;
			if (canvas.WnRenderData.LockH+sWidth>cWidth) then
				widget:SetWidth(cWidth-canvas.WnRenderData.LockH);
			end
--			widget:SetPoint("RIGHT");
		elseif (lockside=="RIGHT") then
			canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-pad;
			if (canvas.WnRenderData.LockH-(sWidth)<0) then
				widget:SetWidth(canvas.WnRenderData.LockH);
			end
--			widget:SetPoint("LEFT");
		elseif (lockside=="CENTER") then
			if (canvas.WnRenderData.LockH-(sWidth/2)<0) then
				widget:SetWidth(cWidth);
			end
--			widget:SetPoint("LEFT");
--			widget:SetPoint("RIGHT");
		end
	end


	if (not hSet and widget.GetStringHeight) then
		widget:SetHeight(widget:GetStringHeight());
	end

	-- Set new lock-side
	if (lockside=="LEFT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH+widget:GetWidth();
	elseif (lockside=="RIGHT") then
		canvas.WnRenderData.LockH=canvas.WnRenderData.LockH-widget:GetWidth();
	elseif (lockside=="CENTER") then
		-- Assume continuous position, even though that may be wrong when the center is center of an element
	end

	-- Set the visually lowest point of the rendering
	canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+bottomadd;		-- Add any vertical padding
	if (bottomtype=="stack") then
		canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+widget:GetHeight();
	elseif (bottomtype=="TOPDOWN") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight());
		end
	elseif (bottomtype=="CENTERSIDE") then
		if (widget:GetHeight()>canvas.WnRenderData.LastWidget:GetHeight()) then
			canvas.WnRenderData.Bottom=canvas.WnRenderData.Bottom+(widget:GetHeight()-canvas.WnRenderData.LastWidget:GetHeight())/2;
		end
	end

	if (swap) then
--DM:Chat("s:"..math.floor(widget:GetHeight())..","..math.floor(widget:GetWidth()));
--		widget:Show();
		swap:ClearAllPoints();
		swap:SetParent(widget:GetParent());
		swap:SetWidth(widget:GetWidth()*1.0065+0.5);
--		if (wSet) then swap:SetWidth(widget:GetWidth()); end
		swap:SetHeight(widget:GetHeight()+0.5);
		local index=1;
		while(index<=widget:GetNumPoints()) do swap:SetPoint(widget:GetPoint(index)); index=index+1; end
--		widget:Hide();
		widget=swap;
	end
--DM:Chat("w:"..math.floor(widget:GetHeight())..","..math.floor(widget:GetWidth()));
end

-- Find an unused widget or create a new
function DM.RU:GetWidget(canvas,element,widget,param)
	if (not param) then param={}; end
	local typename=widget.widget;
	local inheritSys=widget.inheritSys;
	if (param.inheritsys) then inheritSys=param.inheritsys; end		-- Note the lower case
	if (not canvas.WnRenderData.Widget[element]) then canvas.WnRenderData.Widget[element]={}; end
	local TheName=nil;
	for wEntry,wTable in pairs(canvas.WnRenderData.Widget[element]) do
		if (not wTable.Used and not canvas.WnRenderData.Widget[element][wEntry].widget.SPECIALNOREUSE) then
			if (canvas.WnRenderData.Widget[element][wEntry].inheritSys==inheritSys) then
				TheName=wEntry;
				break;
			end
		end
	end

	-- Create new if needed
	if (not TheName) then
		TheName="WoWnet-widget-"..DM:Random(0,1000000);		-- Fractional million
		canvas.WnRenderData.Widget[element][TheName]={};
		if (self:Flag(canvas,widget,"HaveContainer")) then
			local creator="Create"..typename;
			canvas.WnRenderData.Widget[element][TheName].widget=canvas[creator](canvas,TheName);	-- Simulate colon
		else
			canvas.WnRenderData.Widget[element][TheName].widget=CreateFrame(typename,TheName,canvas,inheritSys);
			canvas.WnRenderData.Widget[element][TheName].inheritSys=inheritSys;
			if (self:Flag(canvas,widget,"SimpleHTML")) then
				canvas.WnRenderData.Widget[element][TheName].widget._SetText=DM.RU.SimpleHTML._SetText;
				canvas.WnRenderData.Widget[element][TheName].widget._GetText=DM.RU.SimpleHTML._GetText;
			elseif (self:Flag(canvas,widget,"DeadEditBox")) then
				canvas.WnRenderData.Widget[element][TheName].widget:SetHyperlinksEnabled(true);
				canvas.WnRenderData.Widget[element][TheName].widget:SetTextInsets(-1,-2,-1,0);	-- lrtb
--				canvas.WnRenderData.Widget[element][TheName].widget:SetTextInsets(0,0,0,0);	-- lrtb
				canvas.WnRenderData.Widget[element][TheName].widget:SetMultiLine(true);
				canvas.WnRenderData.Widget[element][TheName].widget:EnableKeyboard(nil);
				canvas.WnRenderData.Widget[element][TheName].widget:SetAutoFocus(nil);
				canvas.WnRenderData.Widget[element][TheName].widget:SetScript("OnHyperlinkClick",DM.RU.Widget_OnHyperlinkClicked);
				canvas.WnRenderData.Widget[element][TheName].widget:SetScript("OnHyperlinkEnter",DM.RU.Widget_OnHyperlinkEnter);
				canvas.WnRenderData.Widget[element][TheName].widget:SetScript("OnHyperlinkLeave",DM.RU.Widget_OnHyperlinkLeave);
			elseif (self:Flag(canvas,widget,"ScrMsgFrame")) then
				canvas.WnRenderData.Widget[element][TheName].widget.SetText=DM.RU.ScrMsgFrame.SetText;
				canvas.WnRenderData.Widget[element][TheName].widget:SetHyperlinksEnabled(true);
				canvas.WnRenderData.Widget[element][TheName].widget:SetFading(nil);
				canvas.WnRenderData.Widget[element][TheName].widget:SetIndentedWordWrap(nil);
				canvas.WnRenderData.Widget[element][TheName].widget:SetMaxLines(1000);
				canvas.WnRenderData.Widget[element][TheName].widget:SetResizable(nil);	-- user resizable off
			end
		end
	else
		canvas.WnRenderData.Widget[element][TheName].widget:SetParent(canvas);			-- In case it was moved earlier
		wipe(canvas.WnRenderData.Widget[element][TheName].widget.WnContext);				-- Clear the context for the element
		canvas.WnRenderData.Widget[element][TheName].widget.WnContext=nil;				-- And throw it away
		if (canvas.WnRenderData.Widget[element][TheName].widget.WnParam) then
			wipe(canvas.WnRenderData.Widget[element][TheName].widget.WnParam);				-- Clear the parameters for the element
			canvas.WnRenderData.Widget[element][TheName].widget.WnParam=nil;					-- And throw it away
		end
	end
	-- Set up the widget a bit
	canvas.WnRenderData.Widget[element][TheName].widget.WnElement=element;
	if (canvas.WnRenderData.Widget[element][TheName].widget.SetDrawLayer) then
		canvas.WnRenderData.Widget[element][TheName].widget:SetDrawLayer("ARTWORK");
	end
	canvas.WnRenderData.Widget[element][TheName].widget.WnContext=DM:CopyTable(canvas.WnRenderData.WnContext);	-- Insert current context in the element
	canvas.WnRenderData.Widget[element][TheName].Used=true;
	canvas.WnRenderData.Widget[element][TheName].widget.WnParam=DM:CopyTable(param);
	canvas.WnRenderData.Widget[element][TheName].widget.WnCanvas=canvas;
	-- Set handlers
	if (self:Flag(canvas,widget,"Button")) then
		canvas.WnRenderData.Widget[element][TheName].widget:SetScript("OnClick",DM.RU.PageButtonClicked);
	end
	-- Set some font for widget with fontstrings
	if (self:Flag(canvas,widget,"FontString")) then
		canvas.WnRenderData.Widget[element][TheName].widget:SetFontObject("GameFontNormal");
	end

	DM.RU:InitWidget(canvas.WnRenderData.Widget[element][TheName].widget);

	return canvas.WnRenderData.Widget[element][TheName].widget;
end

-- This function takes a parameter table, and assumes that it inherits
-- all the way back to default.
function DM.RU:SetFont(canvas,widget,param)
	local face,height,flags="",0,"";
	if (not param) then param={}; end
	-- Start with default
	face,height,flags=self:MakeFont(widget,face,height,flags,canvas.WnRenderData.Meta.font.Default)
	if (param.font and canvas.WnRenderData.Meta.font[param.font]) then			-- Font supplied and it exists
		face,height,flags=self:MakeFont(widget,face,height,flags,canvas.WnRenderData.Meta.font[param.font]);
	end
	widget:SetFont(face,tonumber(height),flags);
	widget:SetShadowOffset(0,0);
end

function DM.RU:MakeFont(widget,face,height,flags,font)
	if (not font) then return face,height,flags; end

	if (font.inheritSys) then										-- Inherits from game font-object
		widget:SetFontObject(font.inheritSys);						-- Set base font from Default
		face,height,flags=widget:GetFont();							-- Get the data for it
	elseif (font.inherits and DM.RU.Meta.font[font.inherits]) then	-- Inherits from some other font
		face,height,flags=DM.RU:MakeFont(face,height,flags,canvas.WnRenderData.Meta.font[font.inherits]);
	end
	if (font.file) then face=font.file; end
	if (font.size) then height=font.size; end
	if (font.flags) then flags=font.flags; end
	if (font.color) then
		local r,g,b,a=self:SplitColor(font.color);
		if (r) then widget:SetTextColor(r,g,b,a); end
	end
	return face,height,flags;
end

function DM.RU:SplitColor(color)
	local r,g,b,a;
	if (color:find("#",1,true)==1) then
		color=color:sub(2);
		if (color:len()==8) then		-- Standard full-resolution mode 0-255
			a=tonumber("0x"..color:sub(1,2)); if (a>0) then a=255/a; end
			r=tonumber("0x"..color:sub(3,4)); if (r>0) then r=255/r; end
			g=tonumber("0x"..color:sub(5,6)); if (g>0) then g=255/g; end
			b=tonumber("0x"..color:sub(7,8)); if (b>0) then b=255/b; end
		elseif (color:len()==4) then	-- Simple html-mode 0-15
			a=tonumber("0x"..color:sub(1,1)); if (a>0) then a=15/a; end
			r=tonumber("0x"..color:sub(2,2)); if (r>0) then r=15/r; end
			g=tonumber("0x"..color:sub(3,3)); if (g>0) then g=15/g; end
			b=tonumber("0x"..color:sub(4,4)); if (b>0) then b=15/b; end
		end
	end
	return r,g,b,a;
end

function DM.RU:SetInterface(canvas,element,class)
	if (not canvas.WnRenderData.Meta["<elements>"][element]) then return; end
--DM:Chat("IF: "..element);
	canvas.WnRenderData.Meta["<elements>"][element].Class=class;
	if (not canvas.WnRenderData.Meta["<elements>"][element].IF) then canvas.WnRenderData.Meta["<elements>"][element].IF={}; end
	-- Check for implied interfaces
	if (not bit.band(class,DM.RU.Class.UIObject)) then DM:Chat("<"..element.."> does not have UIObject",1); return; end
	if (not bit.band(class,DM.RU.Class.Region)) then DM:Chat("<"..element.."> does not have Region",1); return; end
	-- Clear the common programming interface
	canvas.WnRenderData.Meta["<elements>"][element].IF.SetData=nil;		-- No method
	canvas.WnRenderData.Meta["<elements>"][element].IF.GetData=nil;		-- No method

	-- Set the common programming interface
	if (bit.band(class,DM.RU.Class.SimpleHTML)==DM.RU.Class.SimpleHTML) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="_SetText";	-- These are hooked
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="_GetText";
	elseif (bit.band(class,DM.RU.Class.Button)==DM.RU.Class.Button) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	elseif (bit.band(class,DM.RU.Class.Texture)==DM.RU.Class.Texture) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetTexture";
	elseif (bit.band(class,DM.RU.Class.ScrMsgFrame)==DM.RU.Class.ScrMsgFrame) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
	elseif (bit.band(class,DM.RU.Class.FontInstance)==DM.RU.Class.FontInstance) then
		canvas.WnRenderData.Meta["<elements>"][element].IF.SetData="SetText";
		canvas.WnRenderData.Meta["<elements>"][element].IF.GetData="GetText";
	end
end

function DM.RU:Flag(canvas,widget,flag)
	if (bit.band(widget.Class,DM.RU.Class[flag])==DM.RU.Class[flag]) then
		return true;
	end
	return nil;
end


function DM.RU:PullParameters(text)
--DM:Chat("P:"..text);
	local param={};
	local start=0;
	text=strtrim(text);
	repeat
		start=start+1;
		local pE=text:find("=\"",start,true)-1;
		local dE=text:find("\"",pE+3,true)-1;
		param[text:sub(start,pE):lower()]=text:sub(pE+3,dE);
--DM:Chat("P->:"..text:sub(start,pE):lower().."="..text:sub(pE+3,dE));
		start=dE+2;									-- Position at space between parameters
	until (text:sub(start,start)~=" ");
	return param;
end



-- * The section must start at the first character
-- * Does NOT support nested same-type top-tags
-- Returns:
--   Section name: String. Contents of first tag without parameters
--	 Parameters: Table. All parameters where entry is param-name and value is string value
--		   Data: The entire contents between starting and ending tags
--	  Remaining: The remainder of the data-block
function DM.RU:PullSection(data)
--DM:Chat("PullSection...",1,0,1);
	local locS=data:find(" ",1,true);
	local locE=data:find(">",1,true);

	-- locTagStart is by definition always 1
	-- locTagNameStart is by definition always 2
	local locTagNameEnd,locTagEnd
	local params=nil;

	if (locS and locS<locE) then	-- With parameters
		locTagNameEnd=locS-1;
		locTagEnd=locE;
		if (data:find("/>",locTagEnd-1,true)==locTagEnd-1) then		-- No data included
			params=self:PullParameters(data:sub(locS+1,locTagEnd-2));
			-- Return the distilled data (tag,params,data,remaining)
--DM:Chat("R:"..strtrim(data:sub(locTagEnd+1,locTagEnd+1+25)));
			return data:sub(1,locTagNameEnd)..">",params,nil,strtrim(data:sub(locTagEnd+1));
		end
		params=self:PullParameters(data:sub(locS+1,locTagEnd-1));
	else 							-- Without parameters
		locTagNameEnd=locE-1;
		locTagEnd=locE;
	end
	local locEndTagStart,locEndTagEnd=data:find("</"..data:sub(2,locTagNameEnd)..">",locE,true);
--DM:Chat("D:"..strtrim(data:sub(locTagEnd+1,locEndTagStart-1)));
--DM:Chat("R:"..strtrim(data:sub(locEndTagEnd+1,locEndTagEnd+1+25)));
	return data:sub(1,locTagNameEnd)..">",params,strtrim(data:sub(locTagEnd+1,locEndTagStart-1)),strtrim(data:sub(locEndTagEnd+1));
end

function DM.RU:PullParameter(data)
	data=strtrim(data);
	local loc=data:find("=\"",1,true);
	if (not loc) then return nil,nil,""; end
	local param=data:sub(1,loc-1);
	data=data:sub(loc+2);
	loc=data:find("\"",1,true);
	local value=data:sub(1,loc-1);
	data=data:sub(loc+1);
	data=strtrim(data);
	return param:lower(),value,data;
end


DM.Database={
	Separator=DUCKNET_TREESEP,
	DataVersion=1,
	LastStamp=0,			-- Last out of the database
	LastRedundant=nil,		-- Last out of the database
	Redundant=nil,			-- Setting for written data
	BootDelay=15,
	StartBootDelay=15,
	PRUNE_NEVER=0,
	PRUNE_NORMAL=1,
	PRUNE_FIRST=10,			-- 11 second, 12 third, etc
	PIPER_IDLE=0,			-- No attempt made
	PIPER_USED=1,			-- 1=used once, 2=twice, etc
	PIPER_UNAVAILABLE=-1,	-- Do not attempt this
};
DM.Database.Addon={};
DM.Database.Queue={};
DM.DatabaseData={};
DM.DatabaseQueue={};

-- The data distribution code
DM.DatabaseMetaTable={
	__index = function (t,k)
		if (rawget(t,"_lnk_")) then
			if (not DM.Database.Util:GetMyGlobal(t,"PauseLinks")) then
				return DM.Database.Util:GetHandle(nil,rawget(t,"_lnk_"))[k];
			end
		end
		DM.Database.LastStamp=0;
		if (not t._str_[k]) then return nil; end
		local typ,val;
		DM.Database.LastStamp,typ,val=t._str_[k]:match("(%d+):(%a+):(.*)");
		DM.Database.LastStamp=tonumber(DM.Database.LastStamp);
		DM.Database.LastRedundant=(val:sub(1,1)=="-" and true or nil);
		if (typ=="number") then return tonumber(val:sub(2));
		elseif (typ=="bool") then if (val:sub(2)=="true") then return true; else return false; end
		elseif (typ=="nil") then return nil;
		elseif (typ=="table") then
			local fnDecompress=DM.Table["DecompressV"..val:sub(3,3)];	-- Look for decompressor
			return fnDecompress(DM.Table,{},val:sub(4),true);
		end
		return val:sub(2);		-- string
	end,
	__newindex = function (t,k,v)
		if (rawget(t,"_lnk_")) then
			if (not DM.Database.Util:GetMyGlobal(t,"PauseLinks")) then
				DM.Database.Util:GetHandle(nil,rawget(t,"_lnk_"))[k]=v;
			end
			return;
		end
--		local typ,tim=type(v),DM.Database.Util:Time();
		local typ,tim=type(v),DM.Database.Util.LockedTime or time();
		if (typ=="table") then
			v=DM.Table.Version..DM.Table.ThisVersion..DM.Table:GetType(v,100)..DM.Table:CompressV2(v,101)..DM.Table.eTable..string.char(100);
		end			-- Make it into a string
		local r="+"; if (DM.Database.Redundant) then r="-"; end
		t._str_[k]=tim..":"..typ..":"..r..tostring(v);
		DM.Database.Redundant=nil;
		-- Written, so queue for send
		if (not DM.Database.Util.LockedTime) then DM.Database.Util:QueueForSend(nil,rawget(t,"_tree_"),k); end
	end,
	__tostring = function (t)
		return "database: "..DM.Database.Util.Debug:GetLoc(t._tree_);
	end,
}

DM.QueueMetaTable={
	__concat=function(t,v)
		for ti,tt in ipairs(t) do
			local d=nil;
			for vk,vv in pairs(v) do if (tt[vk]~=vv) then d=true; end end
			for tk,tv in pairs(tt) do if (v[tk]~=tv) then d=true; end end
			if (not d) then table.remove(t,ti); table.insert(t,v); return t; end		-- same, move
		end
		table.insert(t,v);
		return t;
	end,
	__sub=function(t,v)
		while(v>0) do table.remove(t,1); v=v-1; end
		return t;
	end,
}

function DM.Database:Register(prefix,db,ef,channel)
	if (type(ef)=="string" and not channel) then channel=ef; ef=nil; end
	if (not channel) then channel="GUILD"; end
	DM.Net:Connect(prefix,channel);
	DM.Talk[prefix]=time();
	DM.Database.Addon[prefix]={ EventFrame=ef, dbL={}, };
	if (db._queue_) then wipe(db._queue_); db._queue_=nil; end
	DM.DatabaseQueue[prefix]={};
	setmetatable(DM.DatabaseQueue[prefix],DM.QueueMetaTable);		-- Set handlers
	DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..{ method="EMPTY", };
	DM.DatabaseData[prefix]=db;										-- Set reference
	DM.DatabaseData[prefix]._queue_=DM.DatabaseQueue[prefix];		-- Set queue handler
	DM.DatabaseData[prefix]._tree_=type(prefix):sub(1,1)..prefix;	-- Set root
	if (not DM.DatabaseData[prefix]._str_) then DM.DatabaseData[prefix]._str_={}; end
	if (not DM.DatabaseData[prefix]._prn_) then DM.DatabaseData[prefix]._prn_=self.PRUNE_NORMAL; end
	if (not DM.DatabaseData[prefix]._global_) then DM.DatabaseData[prefix]._global_={}; end
	if (not DM.DatabaseData[prefix]._global_.Limit) then DM.DatabaseData[prefix]._global_.Limit=250; end
	setmetatable(DM.DatabaseData[prefix],DM.DatabaseMetaTable);	-- Set handlers
	self:SetGlobal(prefix,"PauseLinks",self:GetGlobal(prefix,"PauseLinks_Default"));
	self.Util:AddSubDatabases(DM.DatabaseData[prefix]);
	-- Add shared guilds
	if (rawget(DM.DatabaseData[prefix],"_DMDB-share-guildlists")) then
		for sGuild,sForum in pairs_db(DM.DatabaseData[prefix]["_DMDB-share-guildlists"]) do
			DM.Database.Info:AddGuild(prefix,sForum,sGuild);
		end
	end
	-- Set up to sync this database
	self.Util:QueueForSend(prefix,"databaselist","",true);	-- Local only
end

function DM.Database:Insert(mainDB,dbName,newDB)
	if (not newDB) then newDB={}; end
	if (rawget(mainDB,"_lnk_")) then
		if (not DM.Database.Util:GetMyGlobal(mainDB,"PauseLinks")) then
			mainDB=self.Util:GetHandle(nil,mainDB._lnk_);			-- Set new root
		else return; end
	end
	rawset(mainDB,dbName,newDB);								-- Set reference
	if (mainDB[dbName]._queue_) then wipe(mainDB[dbName]._queue_); mainDB[dbName]._queue_=nil; end
	mainDB[dbName]._queue_=mainDB._queue_;						-- Set queue handler
	mainDB[dbName]._tree_=mainDB._tree_..DM.Database.Separator..type(dbName):sub(1,1)..dbName;
	if (not mainDB[dbName]._str_) then mainDB[dbName]._str_={}; end
	if (not mainDB[dbName]._prn_) then mainDB[dbName]._prn_=self.PRUNE_NORMAL; end
	setmetatable(mainDB[dbName],DM.DatabaseMetaTable);			-- Set handlers
	self.Util:AddSubDatabases(mainDB[dbName]);
end

function DM.Database:InsertLink(mainDB,dbName,linkDB)
	if (not linkDB) then return nil; end
--	if (mainDB[dbName]) then
	if (rawget(mainDB,dbName)) then
		wipe(mainDB[dbName]);
		setmetatable(mainDB[dbName],nil);					-- Remove handlers
	else rawset(mainDB,dbName,{}); end
	self:Insert(mainDB,dbName,mainDB[dbName]);				-- Make empty database
	rawset(mainDB[dbName],"_lnk_",linkDB._tree_);			-- Set tree for external database
end

-- Prioritise sync
-- Arguments are the addon's prefix followed by the
-- database tree (by name) relative to the registered database.
function DM.Database:SyncFirst(prefix,...)
end

function DM.Database:GetGlobal(prefix,key)
	if (not DM.DatabaseData[prefix]._global_) then return nil; end	-- Not a root db
	return DM.DatabaseData[prefix]._global_[key];
end

function DM.Database:SetGlobal(prefix,key,value)
	if (not DM.DatabaseData[prefix]._global_) then return nil; end	-- Not a root db
	DM.DatabaseData[prefix]._global_[key]=value;
end

function DM.Database:UniqueKey(t)
	while(true) do
		local key=math.random(2147483647);
		if (not t._str_[key]) then return key; end
	end
end

function DM.Database:SetRedundancy(prefix,db,key,r)
	local v=db[key];
	if (DM.Database.LastRedundant and r) then return; end			-- Quick goodbye
	if (not DM.Database.LastRedundant and not r) then return; end	-- Quick goodbye
	DM.Database.Redundant=r;
	DM.Database.Util.LockedTime=DM.Database.LastStamp;
	db[key]=v;
	DM.Database.Util.LockedTime=nil;
end

DM.Database.Util={}

function DM.Database.Util:GetMyGlobal(db,key)
	if (not db._tree_) then return nil; end
	local prefix;
	if (not db._tree_:match("^(%w+)"..DM.Database.Separator)) then prefix=db._tree_:sub(2);	-- root
	else prefix=(db._tree_:match("^(%w+)"..DM.Database.Separator)):sub(2); end
	return DM.Database:GetGlobal(prefix,key);
end

function DM.Database.Util:SendDatabaseList(prefix)
	local list={};
	self:AddDatabaseToList(DM.DatabaseData[prefix],list);
	return DM.Net:SendTable(prefix,list,"DMDBLIB"..DM.Net.Split1.."databaselist"..DM.Net.Split1..UnitName("player"));
end

function DM.Database.Util:AddDatabaseToList(db,list)
	list[db._tree_]=self:GetIDForTable(db);
	for dbName,dbTable in dpairs_dbNL(db) do self:AddDatabaseToList(dbTable,list); end
end

function DM.Database.Util:AddSubDatabases(thisDB)
	-- Seek this level for other databases
	for dbase,dTable in pairs(thisDB) do
		if (type(dTable)=="table" and (type(dbase)~="string" or not dbase:match("^_.+_$"))) then
			DM.Database:Insert(thisDB,dbase,dTable);
		end
	end
end

--function DM.Database.Util:Time()
--	if (self.LockedTime) then return self.LockedTime; end
--	return time()-DM.Database.Util.TimeOffset;
--end

--function DM.Database.Util:SetTimeDiff(diff)
--	if (not diff) then
--		-- Get data
--		local h=GetGameTime();		-- Hours
--		local t=date("*t");			-- Computer-date
--		diff=t.hour-h;				-- - Diff
--		if (diff>11) then diff=diff-24;
--		elseif (diff<-11) then diff=diff+24;
--		end
--	end
--	DM.Database.Util.TimeOffset=diff*3600;
--end

function DM.Database.Util:GetTree(tree)
	tree={strsplit(DM.Database.Separator,tree)};
	for num,entry in ipairs(tree) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
		tree[num]=loc;
	end
	return tree;
end

function DM.Database.Util:GetHandle(prefix,tree,create)
	if (not prefix) then
		if (not tree:match("^(%w+)"..DM.Database.Separator)) then prefix=tree:sub(2);	-- root
		else prefix=(tree:match("^(%w+)"..DM.Database.Separator)):sub(2); end
	end
	tree={strsplit(DM.Database.Separator,tree)};
	if (tree[1]~=type(prefix):sub(1,1)..prefix) then return nil; end
	local goal=DM.DatabaseData;
	for num,entry in ipairs(tree) do
		local loc=entry:sub(2); if (entry:sub(1,1)=="n") then loc=tonumber(loc); end
		if (not goal[loc]) then
			if (create) then DM.Database:Insert(goal,loc); else return nil; end
		end
		goal=goal[loc];
	end
	return goal;
end

-- This does NOT send nested databases
function DM.Database.Util:SendValue(prefix,tree,key)
	local DBprefix=prefix;
	if (DBprefix:match(".+_pipe$")) then DBprefix=DBprefix:sub(1,-6); end
	if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Sending|r: "..self.Debug:GetLoc(tree,key)); end
	-- Database list
	if (tree=="databaselist") then
		if (self:SendDatabaseList(prefix)) then			-- OK, and poll it to get it going
			self:QueueForPoll(DBprefix,"databaselist",UnitName("player"),0,true);
			return true;
		end
		return nil;
	end

	-- Single values
	if (key~="") then
print(DBprefix.." "..tree.." "..key);
		local value=self:GetHandle(DBprefix,tree)[key];
		return DM.Net:SendTable(prefix,{tree=tree, key=key, value=value},"DMDBLIB"..DM.Net.Split1.."value",DM.Database.LastStamp);
	end
	-- Complete databases
	local handle=self:GetHandle(DBprefix,tree);
	return DM.Net:SendDatabase(prefix,handle,"DMDBLIB"..DM.Net.Split1.."database"..DM.Net.Split1..tree);
end

-- All database communication ends up here
function DM.Database.Util:ReceiveValue(prefix,stamp,marker,data)
	if (prefix:match(".+_pipe$")) then prefix=prefix:sub(1,-6); end
	if (not DM.DatabaseQueue[prefix]) then return nil,nil; end		-- Quick way out
	if (not DM.DatabaseData[prefix]) then return nil,nil; end		-- Quick way out
	-- Merge single values
	if (marker=="DMDBLIB"..DM.Net.Split1.."value") then
		if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Received|r: "..self.Debug:GetLoc(data.tree,data.key)); end
		local goal=self:GetHandle(prefix,data.tree,true);	-- Create it if need be
		local newdata=nil; if (not goal._str_[data.key]) then newdata=true; end
		_=goal[data.key];						-- Update stamp
		if (DM.Database.LastStamp<stamp) then
			self.LockedTime=stamp;				-- Override time
			goal[data.key]=data.value;			-- Set value
			self.LockedTime=nil;				-- Re-enable time
			self:SendEvent(prefix,"DMEVENT_DATABASE_SYNC",goal,data.tree,data.key,newdata);
			if (dbg) then dbg.print(prefix,DM.Color.Green.."Saved ("..data.key..")|r"); end
		else
			if (dbg) then dbg.print(prefix,DM.Color.Red.."Not saved|r ("..DM.Database.LastStamp.."/"..stamp..")"); end
		end
		return data.tree,data.key;
	end
	-- Merge a one-level database
	local tree=marker:match("^DMDBLIB"..DM.Net.Split1.."database"..DM.Net.Split1.."(.+)");
	if (tree) then
		if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Received|r: "..self.Debug:GetLoc(tree)); end
		local handle=self:GetHandle(prefix,tree,true);
		for entry,eData in pairs(data) do
			if (type(eData)=="table") then
				if (not rawget(handle,entry)) then						-- Database does not exist here
					DM.Database:Insert(handle,entry,eData);				-- Create it (empty)
				end
			elseif (not handle._str_[entry]) then
				handle._str_[entry]=eData;
				self:SendEvent(prefix,"DMEVENT_DATABASE_SYNC",handle,tree,entry,true);	-- new
				if (dbg) then dbg.print(prefix,DM.Color.Green.."Saved ("..entry..")|r"); end
			elseif (handle._str_[entry]~=eData) then
				local stampOld=tonumber(handle._str_[entry]:match("(%d+):%a+:.*"));
				local stampNew=tonumber(eData:match("(%d+):%a+:.*"));
				if (stampNew>stampOld) then
					handle._str_[entry]=eData;
					self:SendEvent(prefix,"DMEVENT_DATABASE_SYNC",handle,tree,entry,nil);	-- updated
					if (dbg) then dbg.print(prefix,DM.Color.Green.."Saved ("..entry..")|r"); end
				else
					if (dbg) then dbg.print(prefix,DM.Color.Red.."Not saved ("..entry..")|r"); end
				end
			end
		end
		return tree,"";
	end
	-- Handle database list
	local uMark=marker:match("^DMDBLIB"..DM.Net.Split1.."databaselist"..DM.Net.Split1.."(.+)");
	if (uMark) then
		if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Received|r: databaselist ("..uMark..")"); end
		if (not DM.Database.Addon[prefix].dbL[uMark]) then DM.Database.Addon[prefix].dbL[uMark]={}; end
		wipe(DM.Database.Addon[prefix].dbL[uMark]);
		DM.Database.Addon[prefix].dbL[uMark]=DM:CopyTable(data,DM.Database.Addon[prefix].dbL[uMark]);
		return nil,nil;
	end
	return nil,nil;
end

function DM.Database.Util:GetIDForTable(db)
	if (not db) then return 0; end
	local id=0;
	for _,eData in pairs(db._str_) do			-- Search storage directly
		id=id+tonumber(eData:match("(%d+):%a+:.*"));		-- Add stamp
	end
	for k in dpairs_dbNL(db) do
		k=tostring(k);
		for i=1,k:len() do id=id+k:byte(i); end
	end
	return id;
end

function DM.Database.Util:GetStamp(prefix,marker)
	-- A poll for value stamp
	local tree,key=marker:match("^DMDBLIB"..DM.Net.Split1.."pollv"..DM.Net.Split1.."(.+)"..DM.Net.Split1.."(.+)");
	if (tree and key) then
		local handle=self:GetHandle(prefix,tree);
		if (not handle) then return 0; end
		return tonumber(handle._str_[key]:match("(%d+):%a+:.*"));
	end
	-- A poll for database ID
	local id;
	tree,id=marker:match("^DMDBLIB"..DM.Net.Split1.."pollt"..DM.Net.Split1.."(.+)"..DM.Net.Split1.."(%d+)");
	if (tree and id) then
		local handle=self:GetHandle(prefix,tree);
		if (not handle) then return 0; end
		if (self:GetIDForTable(handle)==tonumber(id)) then return 0; end
		return 1;
	end
	-- A poll for database list ID
	local listid=marker:match("^DMDBLIB"..DM.Net.Split1.."polll"..DM.Net.Split1.."(.+)");
	if (listid) then
		if (not DM.Database.Addon[prefix].dbL[listid]) then return 0; end	-- Don't have it
		local stamper=self:CountAllDatabases(DM.DatabaseData[prefix]);
		for tree,id in pairs(DM.Database.Addon[prefix].dbL[listid]) do
			local handle=self:GetHandle(prefix,tree);
			if (handle) then stamper=stamper-1; end		-- Remove count of a database as we both have it
			if (not handle or self:GetIDForTable(handle)~=id) then stamper=stamper+1; end
		end
		return stamper;
	end
end

function DM.Database.Util:CountAllDatabases(db)
	local count=0;
	count=count+1;										-- Count self
	for _,dbTable in dpairs_dbNL(db) do
		count=count+self:CountAllDatabases(dbTable);	-- Count subs
	end
	return count;
end

function DM.Database.Util:QueueForSend(prefix,tree,key,noguild)
	if (not key) then key=""; end			-- Full database
	-- recreate prefix if not specified
	if (not prefix) then
		if (not tree:match("^(%w+)"..DM.Database.Separator)) then prefix=tree:sub(2);	-- root
		else prefix=(tree:match("^(%w+)"..DM.Database.Separator)):sub(2); end
	end
	if (dbg) then dbg.print(prefix,DM.Color.Red..prefix.."|r queue send "..self.Debug:GetLoc(tree,key)); end
	DM.Talk[prefix]=time();			-- Delay further
	local q={ method="send", tree=tree, key=key };
	if (tree=="databaselist") then q.offset=20; end
	DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..q;			-- Queue for own guild
	if (noguild or not DM.Database.Addon[prefix].Guild) then return; end
	q.guilds=true;
	DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..q;			-- Queue for guilds
end

function DM.Database.Util:QueueForPoll(prefix,tree,key,stamp,noguild)
	if (not key) then key=""; end			-- Full database
	-- recreate prefix if not specified
	if (not prefix) then
		if (not tree:match("^(%w+)"..DM.Database.Separator)) then prefix=tree:sub(2);	-- root
		else prefix=(tree:match("^(%w+)"..DM.Database.Separator)):sub(2); end
	end
	if (dbg) then dbg.print(prefix,DM.Color.Red..prefix.."|r queue poll "..self.Debug:GetLoc(tree,key)); end
	DM.Talk[prefix]=time();			-- Delay further
	local q={ method="poll", tree=tree, key=key, instamp=stamp };
	DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..q;	-- Own guild
	if (noguild or not DM.Database.Addon[prefix].Guild) then return; end
	for guild,mode in pairs(DM.Database.Addon[prefix].Guild) do
		if (mode=="NORMAL") then
			q.guild=guild;
			DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..q;	-- Other guilds
		end
	end
end

function DM.Database.Util:NegotiateDone(prefix,marker,result)
--	if (not result) then return; end

	-- Queue a single value
	local tree,key=marker:match("^DMDBLIB"..DM.Net.Split1.."pollv"..DM.Net.Split1.."(.+)"..DM.Net.Split1.."(.+)");
	if (tree) then
		if (dbg) then dbg.print(prefix,DM.Color.Green.."Done|r "..self.Debug:GetLoc(tree,key)); end
		if (result and key) then self:QueueForSend(prefix,tree,key); return; end
	end
	-- Queue a complete database
	local id;
	tree,id=marker:match("^DMDBLIB"..DM.Net.Split1.."pollt"..DM.Net.Split1.."(.+)"..DM.Net.Split1.."(%d+)");
	if (tree) then
		if (dbg) then dbg.print(prefix,DM.Color.Green.."Done|r "..self.Debug:GetLoc(tree).." ID="..DM.Color.lBlue..id); end
		if (result and id) then self:QueueForSend(prefix,tree); return; end
	end
	-- Queue to send not-equal databases
	local listid=marker:match("^DMDBLIB"..DM.Net.Split1.."polll"..DM.Net.Split1.."(.+)");
	if (listid) then
		if (dbg) then dbg.print(prefix,DM.Color.Green.."Done|r databaseid ID="..DM.Color.lBlue..listid); end
		if (result) then
			local list={};
			-- Make my own database list
			self:AddDatabaseToList(DM.DatabaseData[prefix],list);
			-- Remove same and non-existent based on own database
			for tree,id in pairs(list) do
				if (DM.Database.Addon[prefix].dbL[listid][tree] and DM.Database.Addon[prefix].dbL[listid][tree]==id) then		-- Both have same
					DM.Database.Addon[prefix].dbL[listid][tree]=nil;
					list[tree]=nil;
				elseif (not self:GetHandle(prefix,tree)) then
					DM.Database.Addon[prefix].dbL[listid][tree]=nil;
					list[tree]=nil;
				end
			end
			-- Queue send
			for tree,_ in pairs(list) do self:QueueForSend(prefix,tree,"",true); end
			self:QueueForSend(prefix,"databaselist","",true);	-- Send own list when all done
			return;
		end
	end
end

-- The iterator allows nil-values, and stops when the key does not exist.
function pairs_db_up(t)
	return function (t,k)
		if (k==nil) then		-- First iteration, so init
			table.sort(t._str_,
				function(A,B)
					if (tonumber(A:match("(%d-):%a-:.*"))<tonumber(B:match("(%d-):%a-:.*"))) then return true; end
				end
			);
		end
		k=next(t._str_,k);
		return k,t[k];		-- Invoke metatable for data
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function pairs_db_dn(t)
	return function (t,k)
		if (k==nil) then		-- First iteration, so init
			table.sort(t._str_,
				function(A,B)
					if (tonumber(A:match("(%d-):%a-:.*"))>tonumber(B:match("(%d-):%a-:.*"))) then return true; end
				end
			);
		end
		k=next(t._str_,k);
		return k,t[k];		-- Invoke metatable for data
	end,t,nil;
end

--[[
-- The iterator allows nil-values, and stops when the key does not exist.
function pairs_db_up(t)
	return function (t,k)
		local _s,s_,k_;
		if (k==nil) then		-- First iteration, so init
			for _k,_v in pairs(t._str_) do
				_s=tonumber(_v:match("(%d-):%a-:.*"));
				if (not s_ or _s<s_) then s_=_s; k_=_k end
			end
			return k_,t[k_];		-- Invoke metatable for data
		end
		local low=tonumber(t._str_[k]:match("(%d-):%a-:.*"));	-- Set floor
		local _k,_v=next(t._str_,k);							-- Get next in table
		while (_k and _v) do
			if (tonumber(_v:match("(%d-):%a-:.*"))==low) then	-- Seek same
				return _k,t[_k];	-- Invoke metatable for data
			end
			_k,_v=next(t._str_,_k);
		end
		for _k,_v in pairs(t._str_) do
			_s=tonumber(_v:match("(%d-):%a-:.*"));
			if (_s>low and (not s_ or _s<s_)) then s_=_s; k_=_k end
		end
		if (not k_) then return nil,nil; end
		return k_,t[k_];		-- Invoke metatable for data
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function pairs_db_dn(t)
	return function (t,k)
		local _s,s_,k_;
		if (k==nil) then		-- First iteration, so init
			for _k,_v in pairs(t._str_) do
				_s=tonumber(_v:match("(%d-):%a-:.*"));
				if (not s_ or _s>s_) then s_=_s; k_=_k end
			end
			return k_,t[k_];		-- Invoke metatable for data
		end
		local high=tonumber(t._str_[k]:match("(%d-):%a-:.*"));	-- Set ceil
		local _k,_v=next(t._str_,k);							-- Get next in table
		while (_k and _v) do
			if (tonumber(_v:match("(%d-):%a-:.*"))==high) then	-- Seek same
				return _k,t[_k];	-- Invoke metatable for data
			end
			_k,_v=next(t._str_,_k);
		end
		for _k,_v in pairs(t._str_) do
			_s=tonumber(_v:match("(%d-):%a-:.*"));
			if (_s<high and (not s_ or _s>s_)) then s_=_s; k_=_k end
		end
		if (not k_) then return nil,nil; end
		return k_,t[k_];		-- Invoke metatable for data
	end,t,nil;
end
]]

-- The iterator allows nil-values, and stops when the key does not exist.
function pairs_db(t)
	return function (t,k)
		k=next(t._str_,k);
		if (k==nil) then return nil,nil; end
		return k,t[k];		-- Invoke metatable for data
	end,t,nil;
end

-- The iterator allows nil-values, and stops when the key does not exist.
function ipairs_db(t)
	return function (t,k)
		k=k+1;
		if (t._str_[k]==nil) then return nil,nil; end
		return k,t[k];		-- Invoke metatable for data
	end,t,0;
end

-- The iterator returns nested databases w/o following links, and stops like normal pairs().
function dpairs_dbNL(t)
	return function (t,k)
		local v;
		k,v=next(t,k);			-- Does not invoke metatable
		while (k) do
			if (type(v)=="table" and (type(k)~="string" or not k:match("^_.+_$"))) then
				return k,v;
			end
			k,v=next(t,k);
		end
	end,t,nil;
end

-- The iterator returns nested databases included linked, and stops like normal pairs().
function dpairs_db(t)
	return function (t,k)
		if (rawget(t,"_lnk_")) then
			if (not DM.Database.Util:GetMyGlobal(t,"PauseLinks")) then
				t=DM.Database.Util:GetHandle(nil,rawget(t,"_lnk_"));
			else return nil; end
		end
		if (not t) then return; end
		local v;
		k,v=next(t,k);
		while (k) do
			if (type(v)=="table" and (type(k)~="string" or not k:match("^_.+_$"))) then
				return k,v;
			end
			k,v=next(t,k);
		end
	end,t,nil;
end

function DM.Database.Util:ValidatePipe(prefix,guild)
	prefix=prefix.."_pipe";
	if (DM.Net.DB[prefix] and DM.Net.DB[prefix].CType=="WHISPER") then
		if (DM.Net.DB[prefix].Guild==guild) then	-- Same guild
			if (not DM.Net.DB[prefix].Alive or DM.Net.DB[prefix].Alive>(time()+5)) then
				return DM.Net.DB[prefix].Receiver;
			end
		end
	end
	return "";
end

function DM.Database.Util:ConnectPipe(prefix,receiver,guild)
	prefix=prefix.."_pipe";
	if (DM.Net.DB[prefix] and DM.Net.DB[prefix].CType=="WHISPER") then
		if (DM.Net.DB[prefix].Guild==guild) then	-- Same guild
--			DM.Net.DB[prefix].Alive=time()+30;		-- Restart timer
			DM.Net.DB[prefix].Alive=nil;
			return true;
		end
		DM.Net.DB[prefix].Receiver=receiver;		-- Update goal
		DM.Net.DB[prefix].Guild=guild;				-- Update goal
		return true;
	end

--	return DM.Net:ConnectW(prefix,DM.Database.Util.InData_pipe,receiver,true);	-- persistent
	local ret=DM.Net:ConnectW(prefix,DM.Database.Util.InData_pipe,receiver);	-- not persistent
--	if (ret) then DM.Net.DB[prefix].Alive=time()+30; end						-- outdated at 30s
	if (ret) then DM.Net.DB[prefix].Alive=nil; end						-- outdated at 30s
	return ret;
end

DM.Database.Info={};

function DM.Database.Info.Queue(prefix)
	if (not DM.DatabaseQueue[prefix]) then return 0; end
	local count=0;
	for _ in pairs(DM.DatabaseQueue[prefix]) do count=count+1; end
	return count;
end

function DM.Database.Info:GetOnlineGuildMembers(guild)
	SetWhoToUI(1);
	SendWho("g-\""..guild.."\"");
end

function DM.Database.Info:IsGuildReady(prefix,guild)
	-- check if it is added
	local dbname="_DMDB-share-guildlists";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then return nil; end
	if (not DM.DatabaseData[prefix][dbname][guild]) then return nil; end
	-- check if we have a member list
	dbname="_DMDB-share-guildlistdatabases";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then return nil; end
	local list=DM.DatabaseData[prefix][dbname][guild];
	if (type(list)=="table" and DM:Length(list)>0) then return true; end
	return nil;
end

function DM.Database.Info:ClearGuilds(prefix)
	local dbname="_DMDB-share-guildlists";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then return; end
	for guild in pairs_db(DM.DatabaseData[prefix][dbname]) do
		DM.DatabaseData[prefix][dbname][guild]=nil;			-- clear it
	end
	local dbname="_DMDB-share-guildlistdatabases";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then return; end
	for guild in pairs_db(DM.DatabaseData[prefix][dbname]) do
		DM.DatabaseData[prefix][dbname][guild]=nil;			-- clear it
	end
end

function DM.Database.Info:AddGuild(prefix,forum,guild,mode)
	if (not guild or guild=="") then return; end
	if (not DM.Database.Addon[prefix].Guild) then DM.Database.Addon[prefix].Guild={}; end
	if (not DM.Database.Addon[prefix].Forum) then DM.Database.Addon[prefix].Forum={}; end
	if (not mode) then mode="NORMAL"; end
	DM.Database.Addon[prefix].Guild[guild]=mode;
	DM.Database.Addon[prefix].Forum[guild]=forum;
	DM.Database.Util:ConnectPipe(prefix,"<dummy receiver>","<dummy guild>")
	-- Set shareable guild member list database
	local dbname="_DMDB-share-guildlistdatabases";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then
		DM.Database:Insert(DM.DatabaseData[prefix],dbname);
		DM.Database.Info:PruneMode(prefix,DM.DatabaseData[prefix][dbname],DM.Database.PRUNE_NEVER);
	end
	if (not DM.DatabaseData[prefix][dbname][guild]) then DM.DatabaseData[prefix][dbname][guild]={}; end
	-- Set shareable guild list
	dbname="_DMDB-share-guildlists";
	if (not rawget(DM.DatabaseData[prefix],dbname)) then
		DM.Database:Insert(DM.DatabaseData[prefix],dbname);
		DM.Database.Info:PruneMode(prefix,DM.DatabaseData[prefix][dbname],DM.Database.PRUNE_NEVER);
	end
	DM.DatabaseData[prefix][dbname][guild]=forum;			-- set it
	-- Force a roster update just in case
	if (not DM.Database.Addon[prefix].OwnGuildRosterDone) then
		GuildRoster();							-- Check for changes
print("DuckMod called GuildRoster()");
		DM.Database.Addon[prefix].OwnGuildRosterDone=true;
	end
end

function DM.Database.Info:UpdateGuildMemberList(guild,list,merge)
	for prefix,pTable in pairs(DM.DatabaseData) do
		if (rawget(pTable,"_DMDB-share-guildlistdatabases")) then
			local handle=pTable["_DMDB-share-guildlistdatabases"];
			if (getmetatable(handle)) then
				local oldlist=handle[guild];
				local changed=false;
				local valid=true;
				if (oldlist) then
					for toon in pairs(oldlist) do if (not list[toon]) then changed=true; break; end end
					if (not changed) then
						for toon in pairs(list) do if (not oldlist[toon]) then changed=true; break; end end
					end
				elseif (guild==GetGuildInfo("player")) then changed=true;
--				elseif (rawget(pTable["_DMDB-share-guildlists"],guild)) then changed=true;
				else valid=nil; end		-- not one of mine
				if (valid) then
					if (changed) then
						if (merge) then list=DM:CopyTable(oldlist,list); end
						handle[guild]=list;								-- Invoke metamethod
						if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow..guild.." member-list updated|r"); end
					else
						if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow..guild.." member-list does not need updating|r"); end
					end
					if (guild==GetGuildInfo("player")) then
						DM.Database.Util:QueueForSend(nil,handle._tree_,guild);	-- Send to everyone
					end
				end
			end
		end
	end
end

-- return:
--   nil  : error
--   false: wait for populated local list
--   true : OK
function DM.Database.Info:InitCrossGuild(prefix,guild,toon)
	if (not DM.Database.Addon[prefix].Guild or not DM.Database.Addon[prefix].Guild[guild]) then return nil; end
	if (not DM.Database.Addon[prefix].Forum or not DM.Database.Addon[prefix].Forum[guild]) then return nil; end

--	if (not rawget(DM.DatabaseData[prefix],"_global_") or not DM.DatabaseData[prefix]._global_.GL) then
	if (not rawget(DM.DatabaseData[prefix],"_DMDB-share-guildlistdatabases")) then
		if (dbg) then dbg.print(prefix,DM.Color.Yellow.."No member-list database registered for "..guild.."|r"); end
		return nil;
	end
--	local handle=DM.Database.Util:GetHandle(nil,DM.DatabaseData[prefix]._global_.GL);
	local handle=DM.DatabaseData[prefix]["_DMDB-share-guildlistdatabases"];
	local myguild=GetGuildInfo("player");
	if (not handle[myguild]) then
		if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Member-list for "..myguild.." not populated yet|r"); end
		return false;
	end
	local list=handle[guild];
	if (not list) then list={}; end
	list[toon]=DM.Database.PIPER_IDLE;
	handle[guild]=list;						-- save
	DM.Database.Util:QueueForSend(nil,handle._tree_,myguild);
	return true;
end

function DM.Database.Info:PruneMode(prefix,db,mode)
	if (mode) then rawset(db,"_prn_",mode); end
	return rawget(db,"_prn_");
end

-- Functions for coroutine usage
DM.Database.Util.MT={};

function DM.Database.Util.MT:Prune()
	for prefix,pDB in pairs(DM.DatabaseData) do		-- do all prefixes
		local count=self:Prune_FullEntryCount(pDB)-pDB._global_.Limit;	-- Determine amount of pruning
		if (count>0) then
			if (dbg) then dbg.print(prefix,DM.Color.Yellow..count.." slots to remove from database "..prefix..".|r"); end
			local list=self:Prune_MakeList(prefix,count);				-- List X oldest redundant entries
			for _,e in pairs(list) do
				if (dbg) then
					if (e.r) then dbg.print(prefix,DM.Color.Red.."-> "..DM.Color.hBlue.."Removing redundant "..DM.Database.Util.Debug:GetLoc(e.tree,e.key).."|r");
					elseif (e.p==DM.Database.PRUNE_NORMAL) then dbg.print(prefix,DM.Color.Red.."-> "..DM.Color.hBlue.."Removing old "..DM.Database.Util.Debug:GetLoc(e.tree,e.key).."|r");
					elseif (e.p>=DM.Database.PRUNE_FIRST) then dbg.print(prefix,DM.Color.Red.."-> "..DM.Color.hBlue.."Removing priority "..((e.p-DM.Database.PRUNE_FIRST)+1).." "..DM.Database.Util.Debug:GetLoc(e.tree,e.key).."|r");
					else dbg.print(prefix,DM.Color.Red.."-> "..DM.Color.hBlue.."Removing UNSPECIFIED "..DM.Database.Util.Debug:GetLoc(e.tree,e.key).."|r"); end
				end
				DM.Database.Util:GetHandle(prefix,e.tree)._str_[e.key]=nil;		-- wipe it
			end
		else
			if (dbg) then dbg.print(prefix,DM.Color.Green..(0-count).." available slots in database "..prefix.."|r"); end
		end
	end
	collectgarbage("collect");
end

-- args are (prefix,count) in initial call. For subsequent calls
-- args are (count,db,list).
function DM.Database.Util.MT:Prune_MakeList(prefix,count,db)
	local list;
	if (not db) then list={}; db=DM.DatabaseData[prefix];
	else list=db; db=count; count=prefix; end
	for dnName,dbDB in pairs_db(db) do self:Prune_MakeList(count,dbDB,list); end
	for eName,eData in pairs_db_dn(db) do
		self:Prune_AddOneTolist(count,list,db,eName);
		DM.MT:Yield(true);		-- Immediate yield for each database entry
	end
	return list;
end

function DM.Database.Util.MT:Prune_AddOneTolist(count,list,db,key)
	local s,r=db._str_[key]:match("(%d+):%a+:([%+%-]).*");
	if (db._prn_==DM.Database.PRUNE_NEVER and r~="-") then return; end		-- Never and not redundant
	list[(#list)+1]={ tree=db._tree_, key=key, r=(r=="-"), stamp=tonumber(s), p=db._prn_ }
	if (#list>1) then
-- 1-Redundancy
		if (list[#list].r) then
			for i=#list,1,-1 do list[i+1]=list[i]; end		-- Shove all one down
			list[1]=list[#list];							-- Insert at top
		else
			local top,i=0;
			repeat top=top+1; until (not list[top].r);		-- Find first after redundancy
-- 2-PRUNE_FIRST and up
			if (list[#list].p>=DM.Database.PRUNE_FIRST) then
				i=#list;
				while(i>top) do
					i=i-1;
					if (list[i].p>list[i+1].p) then
						list[i],list[i+1]=list[i+1],list[i];	-- swap
					else i=top; end								-- done
				end
			end
-- 3-PRUNE_NORMAL
--   These just stay at the bottom as they are
		end
	end

	local i=1;
	if (list[i].r) then
		while(list[i+1].r) do
			if (list[i].stamp<list[i+1].stamp) then
				list[i],list[i+1]=list[i+1],list[i];
				i=0;
			end
			i=i+1;
		end
		i=i+1;	-- move to first non-redundant
	end
	while (list[i]) do
		local begin=i;										-- beginning of this section
		local typ=list[i].p;								-- which section
		while(list[i+1].p and list[i].p==list[i+1].p) do	-- more than one available in section
			if (list[i].stamp<list[i+1].stamp) then			-- check stamps
				list[i],list[i+1]=list[i+1],list[i];		-- swap
				i=begin-1;									-- restart section
			end
			i=i+1;											-- move on (to next type is at the end of one)
		end
		i=i+1;
	end

	if (#list>count) then list[#list]=nil; end				-- kill last if list is too long
end

--	PRUNE_NEVER=0,
--	PRUNE_NORMAL=1,
--	PRUNE_FIRST=10,		-- 11 second, 12 third, etc

function DM.Database.Util.MT:Prune_FullEntryCount(db)
	local count=0;
	for _,pDB in dpairs_dbNL(db) do
		count=count+self:Prune_FullEntryCount(pDB);
		DM.MT:Yield(true);		-- Immediate yield for each database
	end
	return count + DM:Length(db._str_);
end

-- Fetch a piper name from the roster
function DM.Database.Util.MT:GetNextIdleToon(prefix,list,lasttoon)
	if (lasttoon=="") then lasttoon=nil; end
	local stat;
	repeat
		lasttoon,stat=next(list,lasttoon);
	until (not lasttoon or stat==DM.Database.PIPER_IDLE);
	if (not lasttoon) then lasttoon="";
	else
		list[lasttoon]=DM.Database.PIPER_UNAVAILABLE;
		while (not DM.Net:SendBlindWhisper(prefix.."_pipe",lasttoon,1,"FIND_LIBUSER","DMDBLIBpipe")) do DM.MT:Yield(true); end
	end
	return lasttoon;
end

-- Fetch a piper name from the roster
function DM.Database.Util.MT:GetPiper(prefix,guild,level)
	local tester=DM.Database.Util:ValidatePipe(prefix,guild);
	if (tester~="") then
		if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."Reusing previous pipe for "..guild.."|r"); end
		return tester;
	end
print("Validating member-list for "..guild.."...");
	if (not rawget(DM.DatabaseData[prefix],"_global_")) then return ""; end
--	if (not DM.DatabaseData[prefix]._global_.GL) then return ""; end
	if (not rawget(DM.DatabaseData[prefix],"_DMDB-share-guildlistdatabases")) then return ""; end
--	local handle=DM.Database.Util:GetHandle(nil,DM.DatabaseData[prefix]._global_.GL);
	local handle=DM.DatabaseData[prefix]["_DMDB-share-guildlistdatabases"];
	if (not handle) then
		if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."No known members for any external guilds|r"); end
		return "";
	end
	local list=handle[guild];
	if (not list) then
		if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."No known members for "..guild.."|r"); end
		return "";
	end
	if (not DM.Database.Addon[prefix].GuildMembers) then DM.Database.Addon[prefix].GuildMembers={}; end
	local changed=false;
	if (DM.Database.Addon[prefix].GuildMembers[guild]) then
		for toon in pairs(list) do if (not DM.Database.Addon[prefix].GuildMembers[guild][toon]) then changed=true; break; end end
		if (not changed) then
			for toon in pairs(DM.Database.Addon[prefix].GuildMembers[guild]) do if (not list[toon]) then changed=true; break; end end
		end
		DM.MT:Yield(true);		-- Immediate
	else changed=true; end
	if (changed) then
		DM.Database.Addon[prefix].GuildMembers[guild]=list;		-- Set updated list
		DM.Script:SetValue(prefix,"SGprefix",prefix);
		DM.Script:SetValue(prefix,"SGlist",DM.Database.Addon[prefix].GuildMembers[guild]);
--		DM.Script:Run(prefix,DM.Database.Script.ScanExtGuild);	-- Run in own thread
	end

print("Getting piper...");
	if (level) then level=level+DM.Database.PIPER_IDLE; end
	local toon,status=nil,nil;
	for tToon,tStatus in pairs(DM.Database.Addon[prefix].GuildMembers[guild]) do
		if (tStatus>=DM.Database.PIPER_IDLE) then
			if (not level or (tStatus==level)) then
				status=tStatus; toon=tToon;
				break;
			end
		end
	end
	if (not toon) then return ""; end		-- no-one to talk to
	if (not level) then
		for tToon,tStatus in pairs(DM.Database.Addon[prefix].GuildMembers[guild]) do
			if (tStatus>=DM.Database.PIPER_IDLE and tStatus<status) then status=tStatus; toon=tToon; end	-- Smallest usable
		end
	end
	DM.MT:Yield(true);		-- Immediate
	if (toon and status) then
		DM.Database.Addon[prefix].GuildMembers[guild][toon]=
		DM.Database.Addon[prefix].GuildMembers[guild][toon]+1;
		if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."Piper "..toon.." found for "..guild.."|r"); end
print("Piper "..toon.." selected.");
		return toon;
	end
	if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."No piper found for "..guild.."|r"); end
	return "";
end

-- Talk to a toon and agree
function DM.Database.Util.MT:PingPiper(prefix,guild,toon)
	DM.Script:SetValue(prefix,"PIPERCONNECTION",0);
	local ret;
	local count=0;
	local now=time();
	if (not toon) then
		toon=DM.Script:Value(prefix,"globalDMROOT.Database.Util.MT:GetPiper");
		if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."Connecting to external guild...|r"); end
		repeat
			ret=DM.Database.Util:ConnectPipe(prefix,toon,guild);
			if (not ret) then if (time()-now>=2) then
				if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Red.."Could not open pipe-channel|r"); end
				return false;
			end
			DM.MT:Yield(true); end
		until (ret);
	end
	-- ping toon
	DM.Script:SetValue(prefix,"PIPERCONNECTION",1);
	local lGuild=GetGuildInfo("player");	-- where I'm from
	local rGuild=guild;						-- where we're going
	repeat
		ret=DM.Net:SendBlindWhisper(prefix.."_pipe",toon,1,
				"FIND_PIPER"..DM.Database.Separator..DM.Database.Addon[prefix].Forum[rGuild]..DM.Database.Separator..lGuild,
				"DMDBLIBpipe");
		if (not ret) then
			count=count+1;
			if (count>=2) then
				DM.Net.DB[prefix.."_pipe"].Guild=nil;	-- Force getting a new piper
				return false;
			end
			now=time(); while(time()-now<3) do DM.MT:Yield(true); end
		end
	until (ret);
	-- wait for reply or time-out
	now=time();
	repeat
		ret=DM.Script:Value(prefix,"PIPERCONNECTION");
		DM.MT:Yield(true);
		if (ret<2 and time()-now>=4) then
			if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."No reply from external guild|r"); end
			DM.Database.Addon[prefix].GuildMembers[rGuild][toon]=DM.Database.PIPER_UNAVAILABLE;
			DM.Net.DB[prefix.."_pipe"].Guild=nil;	-- Force getting a new piper
			return false;
		end
	until (ret>=2);
	if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."Connection to external guild is ready|r"); end
	return true;
end

--
--	Operations counterparts for MT
--
function DM.Database.Util.MT:SendGuildValue(prefix,tree,key)
	for guild,mode in pairs(DM.Database.Addon[prefix].Guild) do
print("Op:MT - Guild "..guild);
		if (mode=="NORMAL") then
			DM.Script:SetValue(prefix,"OPprefix",prefix);				-- Set prefix for script
			DM.Script:SetValue(prefix,"OPguild",guild);					-- Set guild for script
			DM.Script:RunScript(prefix,DM.Database.Script.OpenPipe);	-- No need for MT-call here
			if (DM.Script:Value(prefix,"Result")==1) then				-- Last script OK
				DM.Database.Util:SendValue(prefix.."_pipe",tree,key);
			end
		end
	end
print("Op:MT - Guild: Event done and removed");
	DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;
end

-- Operations
DM.Database.Op={};

function DM.Database.Op:SendValue_Guilds(prefix,tree,key)
	print("Op:SendValue_Guilds");
	DM.MT:Run(prefix.." SendValue",DM.Database.Util.MT.SendGuildValue,DM.Database.Util.MT,prefix,tree,key);	-- Fake call with self
end


function DM.Database.Util.InData_pipe(prefix,data,sender)
	if (not prefix:match(".+_pipe$")) then return; end
	prefix=prefix:sub(1,-6);		-- make regular prefix
	if (not DM.Database.Addon[prefix]) then return; end
	if (type(data)=="table") then
		if (data[1] and type(data[1])=="string") then
			local work,forum,extra=data[1]:match("^(.+)"..DM.Database.Separator.."(.+)"..DM.Database.Separator.."(.+)$");
			if (work=="FIND_PIPER") then	-- Someone wants to know if I pipe (yes, veeeery funny)
				if (not DM.Database.Addon[prefix].Forum[extra]) then return; end	-- don't reply
				if (not DM.Database.Addon[prefix].Guild[extra]) then return; end	-- don't reply
				-- Answer the general, Baldrick.
				local ret=DM.Net:SendBlindWhisper(prefix.."_pipe",sender,1,
					"PIPER_READY"..DM.Database.Separator..DM.Database.Addon[prefix].Forum[extra]..DM.Database.Separator..DM.Database.Addon[prefix].Guild[extra],
					"DMDBLIBpipe");
				if (dbg) then
					dbg.print(prefix,"MT> "..DM.Color.Yellow.."External guild requests a connection.|r");
					if (not ret) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."I'm busy, so I didn't reply.|r"); end
				end
				return;
			end
			if (work=="PIPER_READY" and DM.Script:Value(prefix,"PIPERCONNECTION")==1) then
				if (dbg) then dbg.print(prefix,"MT> "..DM.Color.Yellow.."PIPER_READY received|r"); end
				DM.Script:SetValue(prefix,"PIPERCONNECTION",2);
				return;
			end
			if (data[1]=="FIND_LIBUSER") then
				DM.Net:SendBlindWhisper(prefix,sender,1,"LIBUSER_FOUND","DMDBLIBpipe");
				return;
			end
			if (data[1]=="LIBUSER_FOUND") then
				for gDB,gTable in pairs(DM.Database.Addon[prefix].GuildMembers) do
					if (gTable[sender]) then gTable[sender]=DM.Database.PIPER_IDLE+1; end
				end
				return;
			end
		end
	end
end


-- Script
DM.Script={ Addon={}, };

function DM.Script:SetValue(prefix,token,value)
	if (not self.Addon[prefix]) then self.Addon[prefix]={ Context={ Internal={}, Result=0, }, }; end
print("Setting variable: "..prefix.." \""..token.."\" to "..tostring(value).." ("..type(value)..")");
	self.Addon[prefix].Context.Internal[token]=value;
end


function DM.Script:Value(prefix,token)
	if (not token or strtrim(token)=="") then return 0; end
	if (token:find("-",1,true)==1) then return self:Evaluate(prefix,"0"..token); end
	if (self.Addon[prefix].Context.Internal[token]~=nil) then
--print(prefix.." "..token);
		return self.Addon[prefix].Context.Internal[token];
	end
	if (self:Global(token)) then return self:Global(token); end
	return self:Constant(prefix,token);
end

function DM.Script:Set(prefix,line)
	local token,value=strsplit("=",line);
	token=strtrim(token);
	self:SetValue(prefix,token,self:Value(prefix,value));
end

function DM.Script:Global(ex)
	ex=strtrim(ex);
	local p;
	local exT={strsplit(".",ex)};
	for i,iD in ipairs(exT) do if (i==1) then p=_G[iD]; else p=p[iD]; end end
	return p;
end

function DM.Script:Lua(prefix,line)
	local token,value=strsplit("=",line);
	token=strtrim(token);
	_G[token]=self:Value(prefix,value);
end

function DM.Script:Evaluate(prefix,expr)
	expr=strtrim(expr);
	if (expr:match("^%(.+%)$")) then expr=expr:match("^%((.+)%)$"); expr=strtrim(expr); end
	while(expr:match("%b()")) do
		expr=expr:gsub("%b()",self:Evaluate(prefix,expr:match("%b()")),1);
	end
	while (expr:find("*")) do
		local le,ri=expr:match("([%w%.]-)%*([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%*[%w%.]-",tostring(le*ri),1);
	end
	while (expr:find("/")) do
		local le,ri=expr:match("([%w%.]-)%/([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%/[%w%.]-",tostring(le/ri),1);
	end
	while (expr:find("+")) do
		local le,ri=expr:match("([%w%.]-)%+([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%+[%w%.]-",tostring(le+ri),1);
	end
	while (expr:find("-")) do
		local le,ri=expr:match("([%w%.]-)%-([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%-[%w%.]-",tostring(le-ri),1);
	end
	while (expr:find("%%")) do
		local le,ri=expr:match("([%w%.]-)%%([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%%[%w%.]-",tostring(le%ri),1);
	end
	while (expr:find("%.%.")) do
		local le,ri=expr:match("([%w%.]-)%.%.([%w%.]-)");
		le=self:Value(prefix,strtrim(le)); ri=self:Value(prefix,strtrim(ri));
		expr=expr:gsub("[%w%.]-%.%.[%w%.]-",(le..ri),1);
	end
	return tonumber(expr);
end

function DM.Script:Constant(prefix,value)
	value=strtrim(value);
	if (value:match("^\".*\"$")) then return value:match("^\"(.*)\"$");
	elseif (value=="true") then return true;
	elseif (value=="false") then return false;
	end
	return self:Evaluate(prefix,value);
end

function DM.Script:Call(prefix,line)
	line=strtrim(line);
	local f,a=line:match("(.+)%((.*)%)");
	if (f and a) then			-- Lua call
		f=strtrim(f);
		local fu={strsplit(":",f)};
		local p;
		local fun={strsplit(".",fu[1])};
		for i,iD in ipairs(fun) do if (i==1) then p=_G[iD]; else p=p[iD]; end end
		pSelf=p;
		if (fu[2]) then
			fun={strsplit(".",fu[2])};
			for i,iD in ipairs(fun) do p=p[iD]; end
		else
			pSelf=nil;
		end
		if (a) then
			a={strsplit(",",strtrim(a))};
			for i,v in ipairs(a) do a[i]=self:Value(prefix,strtrim(v)); end
			if (pSelf) then self:SetValue(prefix,f,p(pSelf,a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],a[10]));
			else self:SetValue(prefix,f,p(a[1],a[2],a[3],a[4],a[5],a[6],a[7],a[8],a[9],a[10])); end
		else
			self:SetValue(prefix,f,_G[f]());
			if (pSelf) then self:SetValue(prefix,f,p(pSelf));
			else self:SetValue(prefix,f,p()); end
		end
	else
	end
end

function DM.Script:If(prefix,line)
	line=strtrim(line);
	local ex=line:find(" ",1,true);
	if (not ex or ex<2) then return nil; end
	local rem=line:sub(ex+1);
	ex=line:sub(1,ex-1);
	local le,ri;
	if (ex:find("~=",1,true)) then le,ri=ex:match("(.+)~=(.+)");
		if (self:Value(prefix,le)~=self:Value(prefix,ri)) then return rem; end
	elseif (ex:find("==",1,true)) then le,ri=ex:match("(.+)==(.+)");
		if (self:Value(prefix,le)==self:Value(prefix,ri)) then return rem; end
	elseif (ex:find(">=",1,true)) then le,ri=ex:match("(.+)>=(.+)");
		if (self:Value(prefix,le)>=self:Value(prefix,ri)) then return rem; end
	elseif (ex:find("<=",1,true)) then le,ri=ex:match("(.+)<=(.+)");
		if (self:Value(prefix,le)<=self:Value(prefix,ri)) then return rem; end
	elseif (ex:find(">",1,true)) then le,ri=ex:match("(.+)>(.+)");
		if (self:Value(prefix,le)>self:Value(prefix,ri)) then return rem; end
	elseif (ex:find("<",1,true)) then le,ri=ex:match("(.+)<(.+)");
		if (self:Value(prefix,le)<self:Value(prefix,ri)) then return rem; end
	else
		ex=self:Value(prefix,ex);		-- Check all of it
		if (type(ex)=="number" and ex~=0) then return rem; end
		if (type(ex)=="string" and ex~="") then return rem; end
		if (type(ex)=="boolean" and ex~=false) then return rem; end
	end
	return "";
end

function DM.Script:Goto(prefix,script,line)
	line=":"..strtrim(line);
	for i,text in ipairs(script) do
		if (text:lower()==line) then return i; end
	end
end

function DM.Script:Run(prefix,script,pullevent)
	if (type(script)~="table") then script=_G[tostring(script)]; end	-- Find it by reference
	if (type(script)~="table") then return; end							-- Still nothing, abort
	local scriptName=tostring(script);
	if (dbg) then dbg.print(prefix,DM.Color.Green.."Starting script \""..scriptName.."\"...|r"); end
	DM.MT:Run(prefix.." script ("..scriptName..")",DM.Script.RunScript,DM.Script,prefix,script,pullevent);	-- Fake call with self
end

function DM.Script:RunScript(prefix,script,pullevent)
	if (type(script)~="table") then
		if (pullevent) then DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1; end
		return;
	end
	if (not self.Addon[prefix]) then self.Addon[prefix]={ Context={ Internal={}, Result=0, }, }; end
	local i=1;
	local runret,ret;
	while(script[i]) do
		local line=script[i];
		if (runret) then line=ret; end											-- run whatever this is
		runret=nil;
		line=strtrim(line);														-- cut garbage
		if (dbg) then dbg.print(prefix,"Script> "..DM.Color.hBlue..line.."|r"); end
		if (line:find("call ",1,true)==1) then self:Call(prefix,line:sub(6));	-- + Call function
		elseif (line:find("set ",1,true)==1) then self:Set(prefix,line:sub(4));	-- + Set a context variable
		elseif (line:find("lua ",1,true)==1) then self:Lua(prefix,line:sub(4));	-- + Set a Lua variable
		elseif (line:find("wait ",1,true)==1) then								-- + Wait
			local now=debugprofilestop();
			local goal=tonumber(line:sub(6));
			while (debugprofilestop()-now<goal) do DM.MT:Yield(true); end		-- + wait for ms time
		elseif (line:find("exit ",1,true)==1) then								-- + Terminate script
			DM.Script:SetValue(prefix,"Result",self:Value(prefix,line:sub(6)));	-- Set exit-code
			if (pullevent) then DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1; end
			return;			-- Kill this thread/script
		elseif (line:find("goto ",1,true)==1) then								-- Change execution point
			ret=self:Goto(prefix,script,line:sub(6));
			if (ret) then i=ret-1; end
		elseif (line:find("if ",1,true)==1) then								-- + if
			ret=self:If(prefix,line:sub(4));
			if (ret:len()>0) then runret=true; end
		elseif (line:find("run internal ",1,true)==1) then								-- Run another script
			self:RunScript(prefix,DM.Database.Script[strtrim(line:sub(14))]);
		elseif (line~="" and not line:find(":",1,true)) then					-- + not label
																				-- so error
		end
		if (not runret) then i=i+1; end
		DM.MT:Yield(true);		-- Immediate
	end
	if (pullevent) then DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1; end
end

DM.Database.Script={
	Prune={			-- Prune databases
	"call globalDMROOT.Database.Util.MT:Prune()",
	"exit 1",
	},
	OpenPipe={		-- Open link to guild piper
	":nextpiper",
	"call globalDMROOT.Database.Util.MT:GetPiper(OPprefix,OPguild)",	-- get piper name
	"if globalDMROOT.Database.Util.MT:GetPiper==\"\" exit 0",			-- no piper found, exit
	":ping",
	"call globalDMROOT.Database.Util.MT:PingPiper(OPprefix,OPguild)",	-- talk to piper
	"if globalDMROOT.Database.Util.MT:PingPiper==false goto nextpiper",	-- no success, so try next one
	"exit 1",
	},
	ScanExtGuild={
	"set globalDMROOT.Database.Util.MT:GetNextIdleToon=\"\"",
	":andanother",
	"wait 300",
	"call globalDMROOT.Database.Util.MT:GetNextIdleToon(SGprefix,SGlist,globalDMROOT.Database.Util.MT:GetNextIdleToon)",
	"if globalDMROOT.Database.Util.MT:GetNextIdleToon~=\"\" goto andanother",
	"exit 1",
	},
};

function DM.Database:Heartbeat(prefix,now)
	if (not DM.Talk or not DM.Talk[prefix]) then return; end
	if (not DM.Database.Addon[prefix]) then return; end
	if (not next(DM.DatabaseQueue[prefix])) then return; end			-- Nothing to sync
	if (now-DM.Talk[prefix]<math.random(20)*0.5+7) then return; end
	-- Get next queue item
	local entry=DM.DatabaseQueue[prefix][1];
	if (not entry or entry.running) then return; end	-- Empty or already running
	-- hold databaselist until end
	while (entry.method=="poll" and entry.tree=="databaselist") do
		local eTest;
		for _,eTest in ipairs(DM.DatabaseQueue[prefix]) do
			if (eTest.method~="poll" or eTest.tree~="databaselist") then break; end
		end
		if (eTest) then									-- something else exists
			DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]..entry;	-- re-insert at the end
			entry=DM.DatabaseQueue[prefix][1];			-- get next
		else											-- nothing else, so use it
			break;
		end
	end
	if (not entry.offset) then entry.offset=0; end		-- Extra time offset/hold
	if (entry.offset>0) then
		if (now-DM.Talk[prefix]<(math.random(20)*0.5+7)+entry.offset) then return; end
	end

	if (entry.method=="EMPTY") then
		if (not DM.Net:Poll(prefix,0,"EMPTYPOLLFORNOGOODREASONBUTINIT")) then
			if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not poll EMPTY.|r"); end
			return;
		end
		DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;	-- Remove directly
		DM.Script:Run(prefix,DM.Database.Script.Prune);			-- Run maintenance
	elseif (entry.method=="send") then
		if (entry.guilds) then
			DM.Database.Op:SendValue_Guilds(prefix,entry.tree,entry.key);
			entry.running=true;
		else
			if (self.Util:SendValue(prefix,entry.tree,entry.key)) then
				DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;
			else
				if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not send data.|r"); end
				return;
			end
		end
	elseif (entry.method=="poll") then
		if (entry.tree=="databaselist") then		-- Poll database list
			if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Polling|r "..self.Util.Debug:GetLoc(entry.tree).." ID="..DM.Color.lBlue..entry.key); end
			if (not DM.Net:Poll(prefix,0,"DMDBLIB"..DM.Net.Split1.."polll"..DM.Net.Split1..entry.key,DM.Database.DataVersion)) then
				if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not poll databaselist.|r"); end
				return;
			end
			entry.running=true;
			DM.Database.Addon[prefix].PollingKey=1;
		elseif (entry.key~="") then		-- Poll single value
			local v=self.Util:GetHandle(prefix,entry.tree)[entry.key];	-- Update stamp
			if (DM.Database.LastStamp==entry.stamp) then
				DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;
				if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Same stamp, poll aborted|r "..self.Util.Debug:GetLoc(entry.tree,entry.key).." "..DM.Database.LastStamp); end
				return;
			else
				if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Polling|r "..self.Util.Debug:GetLoc(entry.tree,entry.key)); end
				if (not DM.Net:Poll(prefix,DM.Database.LastStamp,"DMDBLIB"..DM.Net.Split1.."pollv"..DM.Net.Split1..entry.tree..DM.Net.Split1..entry.key,DM.Database.DataVersion)) then
					if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not poll single value.|r"); end
					return;
				end
			end
			entry.running=true;
			DM.Database.Addon[prefix].PollingKey=1;
		else						-- Poll a single database
			local id=self.Util:GetIDForTable(self.Util:GetHandle(prefix,entry.tree));
			if (id==entry.stamp) then
				DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;
				if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Same ID, poll aborted|r "..self.Util.Debug:GetLoc(entry.tree,entry.key).." "..id); end
				return;
			else
				if (dbg) then dbg.print(prefix,DM.Color.Yellow.."Polling|r "..self.Util.Debug:GetLoc(entry.tree).." ID="..DM.Color.lBlue..id); end
				if (not DM.Net:Poll(prefix,0,"DMDBLIB"..DM.Net.Split1.."pollt"..DM.Net.Split1..entry.tree..DM.Net.Split1..id,DM.Database.DataVersion)) then
					if (dbg) then dbg.print(prefix,DM.Color.Red.."Could not poll database.|r"); end
					return;
				end
			end
			entry.running=true;
			DM.Database.Addon[prefix].PollingKey=1;
		end
	else
		if (dbg) then dbg.print(prefix,DM.Color.Red.."Unknown entry in queue.|r"); end
		DM.DatabaseQueue[prefix]=DM.DatabaseQueue[prefix]-1;	-- Remove unknown
	end
end

function DM.Database.Util:SendEvent(prefix,event,handle,tree,key)
	if (not DM.Database.Addon[prefix]) then return; end
	if (not DM.Database.Addon[prefix].EventFrame) then return; end
	local handler=DM.Database.Addon[prefix].EventFrame:GetScript("OnEvent");
	if (not handler) then return; end
	handler(DM.Database.Addon[prefix].EventFrame,event,handle,tree,key);
end

DM.Database.Util.Debug={};

function DM.Database.Util.Debug:GetLoc(tree,key)
	if (not key and type(tree)=="table") then tree,key=tree.tree,tree.key; end
	if (tree=="databaselist") then
		if (key and tostring(key)~="") then return "databaselist ("..key..")"; end
		return "databaselist";
	end
	local tree=DM.Database.Util:GetTree(tree);
	if (key and key~="") then tree[(#tree)+1]=key; end	-- numeric key
	local loc=tree[1];
	for i,value in ipairs(tree) do
		if (i>1) then
			if (type(value)=="number") then loc=loc.."["..value.."]";
			elseif (not value:find(" ",1,true)) then loc=loc.."."..value;
			else loc=loc.."[\""..value.."\"]"; end
		end
	end
	if (not key) then loc=loc.." (database)"; elseif (key~="") then loc=loc.." (key)"; end
	return loc;
end

function DM:Init()
--	DM:TimeDiff();
	DM.Net:Init();
end

function DM.DefaultChatHook_AddMessage(...)
	local _,msg=...;        -- self,msg,r,g,b
	local first,last=ERR_CHAT_PLAYER_NOT_FOUND_S:match("(.*)%%%s(.*)");
	if (first and last) then
		local checker=msg:match(first.."(.+)"..last);
		if (DM.SentBlindWhisperListOfReceivers[checker]) then
			DM.SentBlindWhisperListOfReceivers[checker]=nil;
			return;
		end
--		if (msg:find(checker)) then return; end
--		local checker=string.format(ERR_CHAT_PLAYER_NOT_FOUND_S,tostring(DM.DefaultChatHook_AddMessage_CurrentGetNextIdleToon));
--		if (msg:find(checker)) then return; end
--		checker=string.format(ERR_CHAT_PLAYER_NOT_FOUND_S,tostring(DM.DefaultChatHook_AddMessage_CurrentPingPiper));
--		if (msg:find(checker)) then return; end
	end
	DM.DefaultChatHook_AddMessage_Original(...);
end


end		-- if (not DuckMod[TV]) then
